# -*- coding: utf-8 -*-
import os
import os.path as op
import string
import cPickle
import tempfile
import logging

import numpy as np
from pkg_resources import Requirement, resource_filename, resource_listdir
import pyhrf
from pyhrf.ndarray import MRI3Daxes, MRI4Daxes, expand_array_in_mask, TIME_AXIS
from nipy.labs import compute_mask_files
from pyhrf.tools import stack_trees, distance
from pyhrf.graph import parcels_to_graphs, kerMask3D_6n, \
    graph_from_mesh, graph_is_sane, sub_graph
from pyhrf.ndarray import xndarray
from pyhrf.xmlio import XmlInitable

try:
    from collections import OrderedDict
except ImportError:
    from pyhrf.tools.backports import OrderedDict


logger = logging.getLogger(__name__)


def _pickle_method(method):
    """
    Allow to safely pickle classmethods. To be fed to copy_reg.pickle
    """
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    if func_name.startswith('__') and not func_name.endswith('__'):
        # deal with mangled names
        cls_name = cls.__name__.lstrip('_')
        func_name = '_%s%s' % (cls_name, func_name)
    return _unpickle_method, (func_name, obj, cls)


def _unpickle_method(func_name, obj, cls):
    """
    Allow to safely unpickle classmethods. To be fed to copy_reg.pickle
    """
    if obj and func_name in obj.__dict__:
        cls, obj = obj, None  # if func_name is classmethod
    for cls in cls.mro():
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)

import copy_reg
import types
copy_reg.pickle(types.MethodType,
                _pickle_method,
                _unpickle_method)


def get_data_file_name(filename):
    """Return the path of a given filename."""
    req = Requirement.parse('pyhrf')
    pyhrf_data_sub_path = 'pyhrf/datafiles'
    filename = os.path.join(pyhrf_data_sub_path, filename)
    return resource_filename(req, filename)


def list_data_file_names():
    """List all the data filenames."""
    req = Requirement.parse('pyhrf')
    pyhrf_data_sub_path = 'pyhrf/datafiles'
    return sorted(resource_listdir(req, pyhrf_data_sub_path))


def get_tmp_path(tag='pyhrf_'):
    """Return a temporary path."""
    tmp_dir = tempfile.mkdtemp(prefix=tag, dir=pyhrf.cfg['global']['tmp_path'])
    return tmp_dir


def get_src_path():
    """Return the source path of pyhrf."""
    return os.path.join(os.path.dirname(pyhrf.__file__), '../../')


def get_src_doc_path():
    """Return the documentation path of pyhrf."""
    return op.join(get_src_path(), 'doc/sphinx/source')


# FIXME: do we need the two following classes?
class AttrClass(object):
    """Base class to display attributes."""

    def __init__(self, **kwargs):
        for k, w in kwargs.iteritems():
            setattr(self, k, w)

    def __repr__(self):
        r = self.__class__.__name__ + '('
        a = [k + '=' + repr(getattr(self, k)) for k in dir(self)
             if not k.startswith('_')]
        r += ','.join(a) + ')'

        return r


class Condition(AttrClass):
    """Represents an activation condition"""
    pass

from pyhrf.tools import PickleableStaticMethod

# PARADIGM STUFFS
from pyhrf.paradigm import Paradigm

DEFAULT_SIMULATION_FILE = get_data_file_name('simu.pck')

# Default localizer onsets
# session-specific onsets here, so onset arrays are not encapsulated
# in lists over all sessions
DEFAULT_ONSETS = OrderedDict(
    [('audio', np.array([15., 20.7, 29.7, 35.4, 44.7, 48., 83.4, 89.7, 108.,
                         119.4, 135., 137.7, 146.7, 173.7, 191.7, 236.7,
                         251.7, 284.4, 293.4, 296.7])),
     ('video', np.array([0., 2.4, 8.7, 33., 39., 41.7, 56.4, 59.7, 75., 96.,
                         122.7, 125.4, 131.4, 140.4, 149.4, 153., 156., 159.,
                         164.4, 167.7, 176.7, 188.4, 195., 198., 201., 203.7,
                         207., 210., 218.7, 221.4, 224.7, 234., 246., 248.4,
                         260.4, 264., 266.7, 269.7, 278.4, 288.]))]
)

DEFAULT_STIM_DURATIONS = OrderedDict(
    [('audio', np.array([])),
     ('video', np.array([]))]
)

fn = 'paradigm_loc_av.csv'
DEFAULT_PARADIGM_CSV = get_data_file_name(fn)

# Default Volumic data
dataFn = 'subj0_bold_session0.nii.gz'
maskFn = 'subj0_parcellation.nii.gz'
DEFAULT_BOLD_VOL_FILE = get_data_file_name(dataFn)
DEFAULT_BOLD_VOL_TR = 2.4
DEFAULT_MASK_VOL_FILE = get_data_file_name(maskFn)
maskFn = 'subj0_parcellation_small.nii.gz'
DEFAULT_MASK_SMALL_VOL_FILE = get_data_file_name(maskFn)
DEFAULT_OUT_MASK_VOL_FILE = './roi_mask.nii'

maskFn = 'real_data_vol_4_regions_mask.nii.gz'
REALISTIC_REAL_DATA_MASK_VOL_FILE = get_data_file_name(maskFn)
boldFn = 'real_data_vol_4_regions_BOLD.nii.gz'
REALISTIC_REAL_DATA_BOLD_VOL_FILE = get_data_file_name(boldFn)

# Default Surfacic data
dataFn = 'real_data_surf_tiny_bold.gii'
maskFn = 'real_data_surf_tiny_parcellation.gii'
meshFn = 'real_data_surf_tiny_mesh.gii'
DEFAULT_BOLD_SURF_FILE = get_data_file_name(dataFn)
DEFAULT_BOLD_SURF_TR = 2.4
DEFAULT_MASK_SURF_FILE = get_data_file_name(maskFn)
DEFAULT_MESH_FILE = get_data_file_name(meshFn)
DEFAULT_OUT_MASK_SURF_FILE = './roiMask.nii'


class FMRISessionVolumicData(XmlInitable):

    parametersComments = {
        'onsets': 'Onsets of experimental simtuli in seconds. \n'
                  'Dictionnary mapping stimulus name to '
                  'the actual list of onsets.',
        'durations': 'Durations of experimental simtuli in seconds.\n'
                     'It has to consistent with the definition of onsets',
        'bold_file': 'Data file containing the 3D+time BOLD signal (nifti format)'
    }

    def __init__(self, onsets=DEFAULT_ONSETS,
                 durations=DEFAULT_STIM_DURATIONS,
                 bold_file=DEFAULT_BOLD_VOL_FILE):

        XmlInitable.__init__(self)
        assert isinstance(onsets, (dict, OrderedDict))
        assert durations is None or isinstance(durations, (dict, OrderedDict))

        self.onsets = onsets
        self.durations = durations
        self.bold_file = bold_file

    def to_dict(self):
        return {'onsets': self.onsets,
                'durations': self.durations,
                'bold_file': self.bold_file}


class FMRISessionSurfacicData(XmlInitable):
    def __init__(self, onsets=DEFAULT_ONSETS,
                 durations=DEFAULT_STIM_DURATIONS,
                 bold_file=DEFAULT_BOLD_SURF_FILE):

        XmlInitable.__init__(self)
        assert isinstance(onsets, dict)
        assert durations is None or isinstance(durations, dict)

        self.onsets = onsets
        self.durations = durations
        self.bold_file = bold_file

    def to_dict(self):
        return {'onsets': self.onsets,
                'durations': self.durations,
                'bold_file': self.bold_file}


class FMRISessionSimulationData(XmlInitable):
    def __init__(self, onsets=DEFAULT_ONSETS,
                 durations=DEFAULT_STIM_DURATIONS,
                 simulation_file=DEFAULT_SIMULATION_FILE):

        XmlInitable.__init__(self)
        assert isinstance(onsets, dict)
        assert durations is None or isinstance(durations, dict)

        self.onsets = onsets
        self.durations = durations
        self.simulation_file = simulation_file

    def to_dict(self):
        return {'onsets': self.onsets,
                'durations': self.durations,
                'bold_file': self.simulation_file}


def load_vol_bold_and_mask(bold_files, mask_file):

    from pyhrf.tools._io import read_volume, discard_bad_data

    # Handle mask
    if not op.exists(mask_file):
        logger.warning('Mask file %s does not exist. Mask is '
                       'computed from BOLD ...', mask_file)

        bold_file = 'subj0_bold_session0.nii.gz'
        # HACK
        if bold_files[0] == pyhrf.get_data_file_name(bold_file):
            max_frac = .99999  # be sure to keep non zero voxels
            connect_component = False  # default BOLD vol has 2 ROIs
        else:
            max_frac = .9
            connect_component = True
        compute_mask_files(bold_files[0], mask_file, False, .4,
                           max_frac, cc=connect_component)
        mask_loaded_from_file = False
    else:
        mask_loaded_from_file = True
        logger.info('Assuming orientation for mask file: ' +
                    string.join(MRI3Daxes, ','))

    logger.info('Read mask from: %s', mask_file)
    mask, mask_meta_obj = read_volume(mask_file)

    if not np.allclose(np.round(mask), mask):
        raise Exception("Mask is not n-ary (%s)" % mask_file)
    mask = np.round(mask).astype(np.int32)

    logger.info('Mask has shape %s\nMask min value: %d\n'
                'Mask max value: %d\nMask has %d parcels',
                str(mask.shape), mask.min(), mask.max(), len(np.unique(mask)))

    if mask.min() == -1:
        mask += 1

    # Load BOLD:
    last_scan = 0
    session_scans = []
    bolds = []
    logger.info('Assuming orientation for BOLD files: ' +
                string.join(MRI4Daxes, ','))

    if type(bold_files[0]) is list:
        bold_files = bold_files[0]

    for bold_file in bold_files:
        if not op.exists(bold_file):
            raise Exception('File not found: ' + bold_file)

        bold, _ = read_volume(bold_file)
        bolds.append(bold)
        session_scans.append(np.arange(last_scan,
                                       last_scan + bold.shape[TIME_AXIS],
                                       dtype=int))
        last_scan += bold.shape[TIME_AXIS]

    bold = np.concatenate(tuple(bolds), axis=TIME_AXIS)

    logger.info('BOLD has shape %s', str(bold.shape))
    discard_bad_data(bold, mask)

    return mask, mask_meta_obj, mask_loaded_from_file, bold, session_scans


def load_surf_bold_mask(bold_files, mesh_file, mask_file=None):
    from pyhrf.tools._io import read_mesh, read_texture, discard_bad_data

    logger.info('Load mesh: ' + mesh_file)
    coords, triangles, coord_sys = read_mesh(mesh_file)
    logger.debug('Build graph...')
    fgraph = graph_from_mesh(triangles)
    assert graph_is_sane(fgraph)

    logger.info('Mesh has %d nodes', len(fgraph))

    logger.debug('Compute length of edges ... ')
    edges_l = np.array([np.array([distance(coords[i],
                                           coords[n],
                                           coord_sys.xform)
                                  for n in nl])
                        for i, nl in enumerate(fgraph)], dtype=object)

    if mask_file is None or not op.exists(mask_file):
        logger.warning('Mask file %s does not exist. Taking '
                       'all nodes ...', mask_file)
        mask = np.ones(len(fgraph))
        mask_meta_obj = None
        mask_loaded_from_file = False
    else:
        mask, mask_meta_obj = read_texture(mask_file)
        mask_loaded_from_file = True

    if not (np.round(mask) == mask).all():
        raise Exception("Mask is not n-ary")

    if len(mask) != len(fgraph):
        raise Exception('Size of mask (%d) is different from size '
                        'of graph (%d)' % (len(mask), len(fgraph)))

    mask = mask.astype(np.int32)
    if mask.min() == -1:
        mask += 1

    # Split graph into rois:
    graphs = {}
    edge_lengths = {}
    for roi_id in np.unique(mask):
        mroi = np.where(mask == roi_id)
        graph, _ = sub_graph(fgraph, mroi[0])
        edge_lengths[roi_id] = edges_l[mroi]
        graphs[roi_id] = graph

    # Load BOLD:
    last_scan = 0
    session_scans = []
    bolds = []
    for bold_file in bold_files:
        logger.info('load bold: %s', bold_file)
        bold, _ = read_texture(bold_file)
        logger.info('bold shape: %s', str(bold.shape))
        bolds.append(bold)
        session_scans.append(np.arange(last_scan, last_scan + bold.shape[0],
                                       dtype=int))
        last_scan += bold.shape[0]

    bold = np.concatenate(tuple(bolds))
    if len(fgraph) != bold.shape[1]:
        raise Exception('Nb positions not consistent between BOLD (%d) '
                        'and mesh (%d)' % (bold.shape[1], len(fgraph)))

    # discard bad data (bold with var=0 and nan values):
    discard_bad_data(bold, mask, time_axis=0)

    return mask, mask_meta_obj, mask_loaded_from_file, bold, session_scans, \
        graphs, edge_lengths


def merge_fmri_sessions(fmri_data_sets):
    """

    fmri_data_sets: list of FmriData objects.
    Each FmriData object is assumed to contain only one session
    """
    from pyhrf.tools import apply_to_leaves

    all_onsets = stack_trees([fd.paradigm.stimOnsets for fd in fmri_data_sets])
    all_onsets = apply_to_leaves(all_onsets, lambda l: [e[0] for e in l])
    durations = stack_trees(
        [fd.paradigm.stimDurations for fd in fmri_data_sets])
    durations = apply_to_leaves(durations, lambda x: [e[0] for e in x])

    bold = np.concatenate([fd.bold for fd in fmri_data_sets])

    session_scan = []
    last_scan = 0
    for fmri_data in fmri_data_sets:
        nscans = fmri_data.bold.shape[0]
        session_scan.append(
            np.arange(last_scan, last_scan + nscans, dtype=int))
        last_scan += nscans

    # FIXME: define fmri_data instead of loop indice
    if fmri_data.simulation is not None:
        simu = [fmri_data.simulation[0] for fmri_data in fmri_data_sets]
    else:
        simu = None

    return FmriData(all_onsets, bold, fmri_data_sets[0].tr, session_scan,
                    fmri_data_sets[0].roiMask, fmri_data_sets[0].graphs,
                    durations, fmri_data_sets[0].meta_obj, simu,
                    backgroundLabel=fmri_data_sets[0].backgroundLabel,
                    data_files=fmri_data_sets[0].data_files,
                    data_type=fmri_data_sets[0].data_type,
                    edge_lengths=fmri_data_sets[0].edge_lengths,
                    mask_loaded_from_file=fmri_data_sets[0].mask_loaded_from_file)



# FIXME: remove the following class and rewrite the next function
class Object(object):
    pass


def merge_fmri_subjects(fmri_data_sets, roiMask, backgroundLabel=0):
    """

    fmri_data_sets: list of FmriData objects, for different subjects.
    In case of multisession data, merging of fmri data over sessions must
    be done for each subject before using this function.
    roiMask: multi_subject parcellation (nparray)
    """

    fmri_subjects = Object()

    np_roi_mask = np.where(roiMask != backgroundLabel)
    roi_ids_in_mask = roiMask[np_roi_mask]
    unique_roi_ids = np.unique(roi_ids_in_mask)

    fmri_subjects.subjects_data = np.append([f for f in fmri_data_sets])
    fmri_subjects.roiMask = roiMask
    fmri_subjects.roiId = unique_roi_ids

    return fmri_subjects


class FmriGroupData(XmlInitable):
    """
    Used for group level hemodynamic analysis
    Encapsulates FmriData objects for all subjects
    All subjects must habe the same number of ROIs

    Inputs:
        list_subjects: contains list of FmriData object for each subject

    """

    def __init__(self, list_subjects):
        self.data_subjects = list_subjects
        self.nbSubj = len(self.data_subjects)

        self.spatial_shape = list_subjects[0].spatial_shape
        self.meta_obj = list_subjects[0].meta_obj

    def getSummary(self, long=False):
        sout = 'Number of subjects: %d\n' % self.nbSubj
        return sout + '\n'.join([ds.getSummary(long=long)
                                 for ds in self.data_subjects])

    def roi_split(self):
        '''
        Retrieve a list of FmriGroupData object,
        each containing the data for all subject, in one ROI
        '''
        rfd = [ds.roi_split() for ds in self.data_subjects]
        nb_rois = len(rfd[0])
        all_rois = []
        for roi_id in xrange(nb_rois):
            rois = []
            for subject_id in xrange(self.nbSubj):
                rois.append(rfd[subject_id][roi_id])
            all_rois.append(FmriGroupData(rois))
        return all_rois

    def build_graphs(self, force=False):
        return [sd.build_graphs(force=force) for sd in self.data_subjects]

    def get_roi_id(self):
        return self.data_subjects[0].get_roi_id()
        # each FmriData object in self.data_subjetcs contains data for
        # all ROIs, for one subject


def get_roi_simulation(simu_sessions, mask, roi_id):
    """
    Extract the ROI from the given simulation dict.
    Args:
        - simu (dict): dictionnary of simulated quantities
        - mask (np.ndarray): binary mask defining the spatial extent of the ROI
        - roi_id (int) : the id of the roi to extract
    Return:
         dict of roi-specific simulation items
    """
    if simu_sessions is None:
        return None

    simus = []
    for simu in simu_sessions:
        roi_simu = {}
        m = np.where(mask == roi_id)

        def duplicate(label):
            if simu.has_key(label):
                roi_simu[label] = simu[label]

        def extract_1D_masked(label):
            if simu.has_key(label):
                roi_simu[label] = simu[label][m[0]]

        def extract_2D_masked(label):
            if simu.has_key(label):
                roi_simu[label] = simu[label][:, m[0]]

        [extract_2D_masked(l) for l in ['labels', 'hrf', 'brf', 'prf', 'nrls',
                                        'prls', 'brls', 'noise', 'drift',
                                        'bold', 'stim_induced_signal',
                                        'drift_coeffs', 'bold_stim_induced',
                                        'perf_stim_induced',
                                        'perf_baseline']]
        #[extract_1D_masked(l) for l in ['perf_baseline']]
        [duplicate(l) for l in ['perf_baseline_var', 'drift_var',
                                'primary_hrf', 'var_subject_hrf',
                                'var_hrf_group', 'condition_defs', 'tr', 'dt',
                                'bold_mixt_params', 'perf_mixt_params']]

        simus.append(roi_simu)
    return simus


class FmriData(XmlInitable):
    """
    Attributes:
        onsets: a dictionary mapping a stimulus name to a list of session onsets. Each item of this list is a 1D numpy 
            float array of onsets for a given session.
        stimDurations: same as 'onsets' but stores durations of stimuli
        roiMask: numpy int array of roi labels (0 stands for the background). Shape depends on the data form 
            (3D volumic or 1D surfacic)
        bold: either a 4D numpy float array with axes [sag,cor,ax,scan] and then
            spatial axes must have the same shape as roiMask, or a 2D numpy float array with axes [scan, position] and 
            position axis must have the same length as the number of positions within roiMask (without background). 
            Sessions are stacked in the scan axis
        sessionsScans: a list of session indexes along scan axis.
        tr: Time of repetition of the BOLD signal
        simulation: if not None then it should be a list of simulation instance.
        meta_obj: extra information associated to data
    """

    parametersComments = {
        'tr': 'repetition time in seconds',
        'sessions_data': 'List of data definition for all sessions',
        'mask_file': 'Input n-ary mask file (= parcellation). '
        'Only positive integers are allowed. \n'
        'All zeros are treated as background positions.'
    }

    parametersToShow = ['tr', 'sessions_data', 'mask_file']

    if pyhrf.__usemode__ == 'devel':
        parametersToShow += ['background_label']

    def __init__(self, onsets, bold, tr, sessionsScans, roiMask, graphs=None,
                 stimDurations=None, meta_obj=None, simulation=None,
                 backgroundLabel=0, data_files=None, data_type=None,
                 edge_lengths=None, mask_loaded_from_file=False,
                 extra_data=None):

        logger.info('Creation of FmriData object ...')
        if isinstance(bold, xndarray):
            logger.info('bold shape: %s', str(bold.data.shape))
        else:
            logger.info('bold shape: %s', str(bold.shape))
        logger.info('roi mask shape: %s', str(roiMask.shape))
        logger.info('unique(mask): %s', str(np.unique(roiMask)))

        sessionsDurations = [len(ss) * tr for ss in sessionsScans]
        self.paradigm = Paradigm(onsets, sessionsDurations,
                                 stimDurations)

        self.tr = tr
        self.sessionsScans = sessionsScans

        self.extra_data = extra_data or {}

        if backgroundLabel is None:
            backgroundLabel = 0

        self.backgroundLabel = backgroundLabel

        logger.info('backgroundLabel: %d', self.backgroundLabel)

        self.store_mask_sparse(roiMask)

        m = self.np_roi_mask
        if isinstance(bold, xndarray):
            if bold.data.ndim > 2:  # volumic data
                print '----------------'
                if data_type is None:
                    data_type = 'volume'
                assert roiMask.ndim == 3
                if TIME_AXIS == 0:
                    bold = bold.data[:, m[0], m[1], m[2]]
                else:  # TIME_AXIS = 3
                    bold = bold.data[m[0], m[1], m[2], :].transpose()

            else:  # surfacic or already flatten data
                assert bold.data.ndim == 2
                if len(m[0]) != bold.data.shape[1]:
                    bold = bold.data[:, m[0]]
                else:
                    bold = bold.data
        else:
            if bold.ndim > 2:  # volumic data
                if data_type is None:
                    data_type = 'volume'
                assert roiMask.ndim == 3
                if TIME_AXIS == 0:
                    bold = bold[:, m[0], m[1], m[2]]
                else:  # TIME_AXIS = 3
                    bold = bold[m[0], m[1], m[2], :].transpose()

            else:  # surfacic or already flatten data
                assert bold.ndim == 2
                if len(m[0]) != bold.shape[1]:
                    bold = bold[:, m[0]]

        self.data_type = data_type

        logger.info('extracted bold shape (without background): %s',
                    str(bold.shape))

        self.bold = bold
        self.bold_full = bold
        self.bold_avg = None

        self.graphs = graphs
        self.edge_lengths = edge_lengths
        self.graphs_full = graphs
        self.graphs_avg = None
        self.meta_obj = meta_obj
        self.simulation = simulation

        self.nbConditions = len(self.paradigm.stimOnsets)
        self.nbSessions = len(sessionsScans)

        self.mask_loaded_from_file = mask_loaded_from_file
        if data_files is None:
            data_files = []
        self.data_files = data_files

    def get_extra_data(self, label, default):
        return self.extra_data.get(label, default)

    def set_extra_data(self, label, value):
        self.extra_data[label] = value

    def get_condition_names(self):
        return self.paradigm.stimOnsets.keys()

    def store_mask_sparse(self, roiMask):

        self.np_roi_mask = np.where(roiMask != self.backgroundLabel)
        self.roi_ids_in_mask = roiMask[self.np_roi_mask]
        self.nb_voxels_in_mask = len(self.roi_ids_in_mask)
        self.spatial_shape = roiMask.shape

    def get_roi_mask(self):
        roi_mask = np.zeros(self.spatial_shape,
                            dtype=self.roi_ids_in_mask.dtype) + \
            self.backgroundLabel
        roi_mask[self.np_roi_mask] = self.roi_ids_in_mask
        return roi_mask
    roiMask = property(get_roi_mask)

    # def __getstate__(self):
    #     return dict((k, v) for (k, v) in self.__dict__.iteritems() \
    #                     if k != 'roiMask')

    def get_nb_rois(self):
        """ Return the number of parcels (background id is discarded) """
        s = set(np.unique(self.roi_ids_in_mask)).discard(0)
        return len(s)

    @PickleableStaticMethod
    def from_vol_files(self, mask_file=DEFAULT_MASK_VOL_FILE,
                       paradigm_csv_file=DEFAULT_PARADIGM_CSV,
                       bold_files=[DEFAULT_BOLD_VOL_FILE],
                       tr=DEFAULT_BOLD_VOL_TR, background_label=None,
                       paradigm_csv_delim=None):
        paradigm = Paradigm.from_csv(paradigm_csv_file,
                                     delim=paradigm_csv_delim)
        durations = paradigm.stimDurations
        onsets = paradigm.stimOnsets

        m, mmo, mlf, b, ss = load_vol_bold_and_mask(bold_files, mask_file)
        mask = m
        mask_meta_obj = mmo
        mask_loaded_from_file = mlf
        bold = b
        sessionScans = ss

        fd = FmriData(onsets, bold, tr, sessionScans, mask,
                      stimDurations=durations, meta_obj=mask_meta_obj,
                      data_files=bold_files + [mask_file, paradigm_csv_file],
                      data_type='volume',
                      mask_loaded_from_file=mask_loaded_from_file,
                      backgroundLabel=background_label)
        fd.set_init(FmriData.from_vol_files, mask_file=mask_file,
                    paradigm_csv_file=paradigm_csv_file,
                    bold_files=bold_files, tr=tr,
                    background_label=background_label,
                    paradigm_csv_delim=paradigm_csv_delim)
        return fd

    @PickleableStaticMethod
    def from_vol_files_rel(self, mask_file, paradigm_csv_file, bold_files,
                           tr, rel_conditions):
        paradigm = Paradigm.from_csv(paradigm_csv_file)
        durations = OrderedDict()
        onsets = OrderedDict()
        for i in xrange(len(rel_conditions)):
            durations[rel_conditions[i]] = paradigm.stimDurations[
                rel_conditions[i]]
            onsets[rel_conditions[i]] = paradigm.stimOnsets[rel_conditions[i]]
        m, mmo, mlf, b, ss = load_vol_bold_and_mask(bold_files, mask_file)
        mask = m
        mask_meta_obj = mmo
        mask_loaded_from_file = mlf
        bold = b
        sessionScans = ss

        fd = FmriData(onsets, bold, tr, sessionScans, mask,
                      stimDurations=durations, meta_obj=mask_meta_obj,
                      data_files=bold_files + [mask_file, paradigm_csv_file],
                      data_type='volume',
                      mask_loaded_from_file=mask_loaded_from_file)
        fd.set_init(FmriData.from_vol_files, mask_file=mask_file,
                    paradigm_csv_file=paradigm_csv_file,
                    bold_files=bold_files, tr=tr)
        return fd

    @PickleableStaticMethod
    def from_vol_ui(self, sessions_data=[FMRISessionVolumicData()],
                    tr=DEFAULT_BOLD_VOL_TR,
                    mask_file=DEFAULT_MASK_VOL_FILE, background_label=None):
        """
        Convenient creation function intended to be used for XML I/O.
        'session_data' is a list of FMRISessionVolumicData objects.
        'tr' is the repetition time.
        'mask_file' is a path to a functional mask file.

        This represents the following hierarchy:
           - FMRIData:
              - list of session data:
                  [ * data for session 1:
                           - onsets for session 1,
                           - durations for session 1,
                           - fmri data file for session 1 (nii)
                    * data for session 2:
                           - onsets for session 2,
                           - durations for session 2,
                           - fmri data file for session 2 (nii)
                  ],
              - repetition time
              - mask file
        """
        logger.info('Load volumic data ...')
        logger.info('Input sessions data:')
        logger.debug(sessions_data)
        sda = stack_trees([sda.to_dict() for sda in sessions_data])
        onsets = sda['onsets']
        durations = sda['durations']
        bold_files = sda['bold_file']

        # FIXME: HACKS!!
        if isinstance(onsets.values()[0][0], list):
            for i in xrange(len(onsets.keys())):
                onsets[onsets.keys()[i]] = onsets[onsets.keys()[i]][0]

        if isinstance(durations.values()[0][0], list):
            for i in xrange(len(durations.keys())):
                durations[durations.keys()[i]] = \
                    durations[durations.keys()[i]][0]

        if len(durations.values()[0][0].shape) > 1:
            for i in xrange(len(durations.keys())):
                durations.values()[i][0] = durations.values()[i][0][0]

        if len(onsets.keys()) == 1:
            onsets[onsets.keys()[0]] = onsets[onsets.keys()[0]][0]
            durations[onsets.keys()[0]] = durations[onsets.keys()[0]][0]
        if isinstance(durations, list) and durations[0] is None:
            durations = None

        mask, mmo, mlf, bold, session_scans = \
            load_vol_bold_and_mask(bold_files, mask_file)
        mask_meta_obj = mmo
        mask_loaded_from_file = mlf
        fmri_data = FmriData(onsets, bold, tr, session_scans, mask,
                             stimDurations=durations, meta_obj=mask_meta_obj,
                             data_files=bold_files + [mask_file],
                             backgroundLabel=background_label,
                             data_type='volume',
                             mask_loaded_from_file=mask_loaded_from_file)

        fmri_data.set_init(FmriData.from_vol_ui, sessions_data=sessions_data,
                           tr=tr, mask_file=mask_file,
                           background_label=background_label)
        return fmri_data

    @classmethod
    def from_surf_files(self, paradigm_csv_file=DEFAULT_PARADIGM_CSV,
                        bold_files=None, tr=DEFAULT_BOLD_SURF_TR,
                        mesh_file=DEFAULT_MESH_FILE, mask_file=None):
        """Return FmriData representation from surf files"""
        if bold_files is None:
            bold_files = [DEFAULT_BOLD_SURF_FILE]

        paradigm = Paradigm.from_csv(paradigm_csv_file)
        durations = paradigm.stimDurations
        onsets = paradigm.stimOnsets

        mask, mask_meta_obj, _, bold, session_scans, graphs, edge_lengths = \
            load_surf_bold_mask(bold_files, mesh_file, mask_file)

        fmri_data = FmriData(onsets, bold, tr, session_scans, mask, graphs,
                             stimDurations=durations, meta_obj=mask_meta_obj,
                             data_files=bold_files + [mask_file, mesh_file],
                             data_type='surface', edge_lengths=edge_lengths)
        fmri_data.set_init(FmriData.from_surf_files,
                           paradigm_csv_file=paradigm_csv_file,
                           bold_files=bold_files, tr=tr, mesh_file=mesh_file,
                           mask_file=mask_file)
        return fmri_data

    @classmethod
    def from_surf_ui(cls, sessions_data=None, tr=DEFAULT_BOLD_SURF_TR,
                     mask_file=DEFAULT_MASK_SURF_FILE,
                     mesh_file=DEFAULT_MESH_FILE):
        """
        Convenient creation function intended to be used for XML I/O.
        'session_data' is a list of FMRISessionVolumicData objects.
        'tr' is the time of repetition.
        'mask_file' is a path to a functional mask file.

        This represents the following hierarchy:
        
        .. code:: 
        
            - FMRIData:
               - list of session data:
                   [ * data for session 1:
                            - onsets for session 1,
                            - durations for session 1,
                            - fmri data file for session 1 (gii)
                     * data for session 2:
                            - onsets for session 2,
                            - durations for session 2,
                            - fmri data file for session 2 (gii)
                   ],
               - time of repetition
               - mask file
               - mesh file
        """

        if sessions_data is None:
            sessions_data = [FMRISessionSurfacicData()]

        logger.info('Load surfacic data...')

        sda = stack_trees([sda.to_dict() for sda in sessions_data])
        onsets = sda['onsets']
        durations = sda['durations']
        bold_files = sda['bold_file']

        if isinstance(durations, list) and durations[0] is None:
            durations = None

        (mask, mask_meta_obj, _, bold, session_scans,
         graphs, edge_lengths) = load_surf_bold_mask(bold_files, mesh_file,
                                                     mask_file)

        fmri_data = FmriData(onsets, bold, tr, session_scans, mask, graphs,
                             stimDurations=durations, meta_obj=mask_meta_obj,
                             data_files=bold_files + [mask_file, mesh_file],
                             data_type='surface', edge_lengths=edge_lengths)
        fmri_data.set_init(FmriData.from_surf_ui, sessions_data=sessions_data,
                           tr=tr, mask_file=mask_file, mesh_file=mesh_file)
        return fmri_data

    @classmethod
    def from_simu_ui(cls, sessions_data=None):
        if sessions_data is None:
            sessions_data = [FMRISessionSimulationData()]
        # load simu.pck -> b
        logger.debug('from_simu_ui\n'
                     'simulation file: %s', sessions_data[0].simulation_file)
        bs = [cPickle.load(file(sd.simulation_file)) for sd in sessions_data]
        # print 'condName:',b.conditions
        # print 'onsets:',b.onsets
        # print 'durations:',b.durations
        # print 'b:',b
        # print 'b type is :', type(b)
        fds = []
        for b in bs:
            fd = FmriData.from_simulation_dict(b)
            fds.append(fd)

        fds = merge_fmri_sessions(fds)

        fds.set_init(FmriData.from_simu_ui, sessions_data=sessions_data)
        return fds

    @classmethod
    def from_simulation_dict(self, simulation, mask=None):
        from pyhrf.tools._io import read_volume

        bold = simulation['bold']

        if isinstance(bold, xndarray):
            bold = bold.reorient(['time', 'voxel'])
            nvox = bold.data.shape[1]
            ss = [range(bold.data.shape[0])]  # one session
            print 'BOLD SHAPE=', bold.data.shape
        else:
            nvox = bold.shape[1]
            ss = [range(bold.shape[0])]  # one session

        onsets = simulation['paradigm'].stimOnsets
        # onsets = dict(zip(ons.keys(),ons.values()]))
        durations = simulation['paradigm'].stimDurations
        #durations = dict(zip(dur.keys(),[o for o in dur.values()]))

        # print ''
        # print 'onsets:'
        # print onsets
        # print ''
        tr = simulation['tr']

        s = simulation
        labelsVol = s.get('labels_vol', np.ones((1, nvox,)))

        if mask is None:
            if isinstance(labelsVol, xndarray):
                default_mshape = labelsVol.data.shape[1:]
            else:
                default_mshape = labelsVol.shape[1:]
                # print 'default_mask_shape:', default_mshape
                roiMask = simulation.get('mask', np.ones(default_mshape,
                                                         dtype=np.int32))
        else:
            roiMask = read_volume(mask)[0]

        # print 'roiMask:', roiMask.shape
        if len(roiMask.shape) == 3:
            data_type = 'volume'
        else:
            data_type = 'surface'  # simulation ??

        return FmriData(onsets, bold, tr, ss, roiMask, stimDurations=durations,
                        simulation=[simulation], data_type=data_type)

    def get_data_files(self):
        return self.data_files

    def save(self, output_dir):
        """
        Save paradigm to output_dir/paradigm.csv,
        BOLD to output_dir/bold.nii, mask to output_dir/mask.nii
        #TODO: handle multi-session

        Return: tuple of file names in this order: (paradigm, bold, mask)
        """
        from pyhrf.tools._io import write_volume, write_texture
        paradigm_file = op.join(output_dir, 'paradigm.csv')
        self.paradigm.save_csv(paradigm_file)
        if self.data_type == 'volume':
            # unflatten bold
            bold_vol = expand_array_in_mask(self.bold, self.roiMask, 1)
            bold_vol = np.rollaxis(bold_vol, 0, 4)
            bold_file = op.join(output_dir, 'bold.nii')
            write_volume(bold_vol, bold_file, self.meta_obj)

            mask_file = op.join(output_dir, 'mask.nii')
            write_volume(self.roiMask, mask_file, self.meta_obj)

        elif self.data_type == 'surface':  # TODO surface
            bold_file = op.join(output_dir, 'bold.gii')
            write_texture(self.bold_vol, bold_file, self.meta_obj)
            pass

        return paradigm_file, bold_file, mask_file

    def build_graphs(self, force=False):
        logger.debug('FmriData.build_graphs (self.graphs is None ? %s ) ...',
                     str(self.graphs is None))
        logger.debug('data_type: %s', self.data_type)
        if self.graphs is None or force:
            if self.data_type == 'volume':
                logger.info('Building graph from volume ...')
                to_discard = [self.backgroundLabel]
                self.graphs = parcels_to_graphs(self.roiMask,
                                                kerMask3D_6n,
                                                toDiscard=to_discard)
                logger.info('Graph built (%d rois)!', len(self.graphs.keys()))
                self.edge_lentghs = dict([(i, [[1] * len(nl) for nl in g])
                                          for i, g in self.graphs.items()])
            elif self.data_type == 'surface':
                if self.graphs is not None:
                    return self.graphs
                else:
                    raise Exception('Graph is not set for surface data!')

    def get_roi_id(self):
        """ In case of FMRI data containing only one
        ROI, return the id of this ROI. If data contains several ROIs
        then raise an exception
        """
        # roi_ids_in_mask = set(self.roiMask.flatten())
        # if self.backgroundLabel in roi_ids_in_mask:
        #     roi_ids_in_mask.remove(self.backgroundLabel)
        # roi_ids_in_mask = list(roi_ids_in_mask)
        unique_roi_ids = np.unique(self.roi_ids_in_mask)
        if len(unique_roi_ids) > 1:
            raise Exception('FMRI data contains more than one ROI')
        return unique_roi_ids[0]

    def get_nb_vox_in_mask(self):
        return self.nb_voxels_in_mask
        # return (self.roiMask != self.backgroundLabel).sum()

    def get_graph(self):
        self.build_graphs(force=True)
        if len(self.graphs) == 1:
            return self.graphs[self.graphs.keys()[0]]
        else:
            return self.graphs

    def roi_split(self, mask=None):
        if mask is None:
            mask = self.roiMask

        assert mask.shape == self.roiMask.shape

        onsets = self.paradigm.stimOnsets
        durations = self.paradigm.stimDurations

        in_mask = mask[np.where(mask != self.backgroundLabel)]
        data_rois = []
        for roiId in np.unique(mask):
            if roiId != self.backgroundLabel:
                mroi = np.where(mask == roiId)
                roiMask = np.zeros_like(mask) + self.backgroundLabel
                roiMask[mroi] = roiId
                mroi_bold = np.where(in_mask == roiId)
                roiBold = self.bold[:, mroi_bold[0]]
                roiGraph = None
                if self.graphs is not None:
                    roiGraph = self.graphs[roiId]
                    logger.info('graph for roi %s has %d nodes',
                                roiId, len(roiGraph))
                else:
                    logger.info('graph for roi %s is None', (roiId))

                simulation = get_roi_simulation(
                    self.simulation, in_mask, roiId)

                data_rois.append(FmriData(onsets, roiBold, self.tr,
                                          self.sessionsScans, roiMask,
                                          {roiId: roiGraph},
                                          durations, self.meta_obj,
                                          simulation, self.backgroundLabel,
                                          self.data_files, self.data_type))
        return data_rois

    def discard_small_rois(self, min_size):
        too_small_rois = [i for i in np.unique(self.roi_ids_in_mask)
                          if (self.roi_ids_in_mask == i).sum() < min_size]
        logger.info(" %d too small ROIs are discarded (size < %d)",
                    len(too_small_rois), min_size)
        self.discard_rois(too_small_rois)

    def discard_rois(self, roi_ids):
        roiMask = self.roiMask
        for roi_id in roi_ids:
            mroi = np.where(roiMask == roi_id)
            roiMask[mroi] = self.backgroundLabel
        self.store_mask_sparse(roiMask)
        if self.graphs is not None:
            self.build_graphs(force=True)

    def keep_only_rois(self, roiIds):
        roiToKeep = set(np.unique(roiIds))
        roiToRemove = set(
            np.unique(self.roi_ids_in_mask)).difference(roiToKeep)
        self.discard_rois(roiToRemove)

    def get_joined_onsets(self):
        return self.paradigm.get_joined_onsets()

    def get_joined_durations(self):
        return self.paradigm.get_joined_durations()

    # TODO : fix average computing -> handle sparse representation of mask
    def compute_average(self):
        # raise NotImplementedError("TODO : fix average computing -> "\
        #                               "handle sparse representation of mask")
        in_mask = self.roiMask[np.where(self.roiMask != self.backgroundLabel)]
        bg_lab = self.backgroundLabel
        roi_ids_in_mask = set(self.roiMask.flatten())
        roi_ids_in_mask.remove(bg_lab)
        self.mask_avg = np.array(sorted(roi_ids_in_mask))
        self.bold_avg = np.zeros((self.bold.shape[0], self.mask_avg.size))
        for i, roi_id in enumerate(self.mask_avg):
            mroi = np.where(in_mask == roi_id)
            self.bold_avg[:, i] = self.bold[:, mroi[0]].mean(1)
        if self.graphs_avg is None and self.graphs is not None:
            # average scenario -> only one position per roi:
            self.graphs_avg = [[]] * len(self.graph)
        self.graph = self.graphs_avg

    def average(self, flag=True):
        # raise NotImplementedError("TODO : fix average computing -> "\
        #                               "handle sparse representation of mask")
        self.average_flag = flag
        if self.average_flag:
            if self.bold_avg is None:
                self.compute_average()

            self.bold = self.bold_avg
            self.graphs = self.graphs_avg
            #self.roiMask = self.mask_avg
        else:
            self.bold = self.bold_full
            self.graphs = self.graphs_full
            #self.roiMask = self.roiMask_full

    def getSummary(self, long=False):
        """
        """
        bg_lab = self.backgroundLabel

        parcel_sizes = np.bincount(self.roiMask.flatten())

        s = ''
        s += self.paradigm.get_info(long=long)
        roiIds = np.unique(self.roiMask)
        s += ' - geometry: %s [%s]\n' % (str(self.roiMask.shape),
                                         string.join(MRI3Daxes, ','))
        s += ' - graph computed: %s\n' % str(self.graphs is not None)
        s += ' - %d rois - [%d' % (len(roiIds), roiIds[0])
        if len(roiIds) > 1:
            s += '...%d]\n' % (sorted(roiIds)[-1])
        else:
            s += ']\n'
        s += ' - background label: %d\n' % (self.backgroundLabel)
        bg_size = (self.roiMask == bg_lab).sum()
        s += ' - background size: %d\n' % bg_size
        s += ' - nb voxels (without background): %d\n' \
            % (self.roiMask.size - bg_size)
        if self.meta_obj is not None:
            if isinstance(self.meta_obj, dict) and \
                    self.meta_obj.has_key('voxel_size'):
                s += ' - voxel size : %s\n' % str(self.meta_obj['voxel_size'])
        biggest_parcel = np.argmax(parcel_sizes)
        # print 'biggest_parcel:', biggest_parcel
        s += ' - biggest parcel : %d (size=%d)\n' \
            % (biggest_parcel, parcel_sizes[biggest_parcel])
        s += ' - nb sessions: %d\n' % self.nbSessions
        s += ' - scans per session:\n'
        for si, ss in enumerate(self.sessionsScans):
            s += '   sess %d -> [%d...%d] (%d)\n' % (si,
                                                     ss[0], ss[-1], len(ss))
        # print 'type tr:', self.tr, type(self.tr)
        s += ' - tr : %1.2f\n' % self.tr
        if long:
            in_mask = self.roiMask[np.where(self.roiMask != bg_lab)]
            # print 'in_mask:', in_mask.size
            # print roiIds
            # print self.bold.keys()
            for i in np.unique(in_mask):
                mroi = np.where(in_mask == i)
                bold_roi = self.bold[:, mroi[0]]
                s += '  .. roid %d: size=%d \n' \
                    % (i, parcel_sizes[i])
                s += '     bold: %1.2f(s%1.2f)[%1.2f;%1.2f]\n' \
                    % (bold_roi.mean(), bold_roi.std(),
                       bold_roi.min(), bold_roi.max())
        return s

    def __repr__(self):
        """
        Return a readable string representation of the object.
        Ensures that if 2 objects are instanciated with the same parameters
        they yield the same representation.
        """
        r = self.__class__.__name__ + '('
        r += 'onsets=' + repr(self.paradigm.stimOnsets) + ','
        r += 'bold=' + repr(self.bold) + ','
        r += 'tr=' + repr(self.tr) + ','
        r += 'sessionsScans=' + repr(self.sessionsScans) + ','
        r += 'roiMask=' + repr(self.roiMask) + ','
        r += 'graph=' + repr(self.graphs) + ','
        r += 'stimDurations=' + repr(self.paradigm.stimDurations) + ','
        r += 'meta_obj=' + repr(self.meta_obj)  # + ','
        #r += 'simulation=' + repr(self.simulation)
        # TODO: make proper 'readable' __repr__ function for BoldModel object
        r += ')'
        return r



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module to distribute shell commands across a network

Original Author: Mathieu Perrot
Extended by: Thomas Vincent
"""

import os, sys, numpy, re, getpass, time, datetime, signal
import exceptions, warnings
import socket, threading
from optparse import OptionParser
warnings.filterwarnings('ignore', 'Python C API version mismatch',
                                                RuntimeWarning)
try:
    import paramiko
except ImportError:
    pass

from pyhrf.tools.message import msg

# Commentaires divers :
# ---------------------
# - Ce script suppose que l'hote distant est un linux, kernel 2.6. A tester avec
# kernel 2.4...pas sur que ca marche : a cause du contenu de /proc.
# - Ce script utilise le fait que le shell distant est un shell bash,
# il faudrait tester avec d'autre shell pour voir s'il marche avec sh.
# - il utilise les commandes distantes suivantes :
#   * cat, sed, grep, tty
# - il utilise les builtins du shell suivantes :
#   * ulimit


# FIXME
# 1) ajouter aide detaillee pour --user, --password
# pour --password : 'GET', '', '-' : pour rentrer le mot de passe a la main
# 2) mode tache hierarchique :
#    - erreur si 2 fois la meme regle de defini
#    - erreur si boucle : au runtime c'est plus facile
#    - je crois que seul les arbres marche.


def remote_dir_is_writable(user, hosts, path):
    """
    Test if *path* is writable from each host in *hosts*. Sending bash
    commands to each host via ssh using the given *user* login.

    Args:

    """
    import os.path as op
    import pyhrf

    mode = 'dispatch'
    cmds = ['bash -c "echo -n \"[write_test]:%d:\"; '\
            'if [ -w %s ]; then echo \"OK\"; else echo \"NOTOK\"; fi;"' \
            %(ih, path) for ih in range(len(hosts))]
    tasks = read_tasks(cmds, mode)
    timeslot = read_timeslot('allday')

    tmp_dir = pyhrf.get_tmp_path()
    logfile = op.join(tmp_dir, 'pyhrf.log')
    run_grid(mode, hosts, 'rsa', tasks, timeslot, logfile=logfile, user=user)
    kill_threads()
    log = open(logfile).readlines()

    res = [False] * len(hosts)
    for line in log:
        if line.startswith('[write_test]'):
            #print line
            _,ih,r = line.strip('\n').split(':')
            res[int(ih)] = ('OK' in r)
    os.remove(logfile)
    return res

def create_options(argv):
    cmd = os.path.basename(argv[0])
    #description = 'Grid computing'
    usage = '''Grid computing
%s --hosts HOSTS --tasks TASKS [options]
   Dispatch TASKS list on HOSTS list.
%s --hosts HOSTS [options]
   Only probe hosts.''' % (cmd, ''.ljust(len(cmd)))
    parser = OptionParser(usage = usage)
    parser.add_option('--hosts', dest='hosts',
        metavar='HOSTS', action='store', default=None,
        help='hosts file or hosts list ("host1, host2, host3")')
    parser.add_option('--morehelp', dest='morehelp',
        metavar='OPT', action='store', default=None,
        help='details help on OPT (ex : --morehelp hosts)')
    parser.add_option('--tasks', dest='tasks',
        metavar='FILE', action='store', default=None,
        help='tasks file (one task by line)')
    parser.add_option('--mode', dest='mode',
        metavar='MODE', action='store', default='dispatch',
        help='task manager mode : dispatch,repeat,hierarchical/hie')
    parser.add_option('--log', dest='logfile',
        metavar='FILE', action='store', default="grid.log",
        help='log output(stdout/stderr) of commands (default %default)')
    parser.add_option('--broken', dest='brokenfile',
        metavar='FILE', action='store', default="broken_cmd.batch",
        help='file to store commands returning bad exist status '
            '(default %default)')
    parser.add_option('--user', dest='user',
        metavar='USER', action='store', default=None,
        help='set user name (default $USER)')
    parser.add_option('--passwd', dest='passwd',
        metavar='PASSWD', action='store', default=None,
        help='set user password (default : no password, use ssh key)')
    parser.add_option('--keytype', dest='keytype',
        metavar='TYPE', action='store', default='dsa',
        help='dsa or rsa (key type)')
    parser.add_option('--timeslot', dest='timeslot',
        metavar='T', action='store', default='night',
        help='define timeslot when tasks can be started (default : night)')
    return parser

def log_help(cmd):
    print '''
--------------------------------------------------------------------------------
 --log FILE (FILE default value = grid.log)

    Store standard output and error output of each task, logging starting date,
    endding date, computing time, computing host and associated command line.
--------------------------------------------------------------------------------
'''

def broken_help(cmd):
    print '''
--------------------------------------------------------------------------------
 --broken FILE (FILE default value = broken_cmd.batch)

    Store all broken commands (those which returns bad exist status, i.e
    different from 0 : see echo $? on bash shell) on FILE.

    $ cat FILE
    ls ~/plop
    $ ls ~/plop
    ls : /home/???/plop: No such file or directory
    $ echo $?
    1
--------------------------------------------------------------------------------
'''

def mode_help(cmd):
    print '''
--------------------------------------------------------------------------------
 --mode MODE
    MODE in [ dispatch, repeat ]

 * dispatch :   each task is dispatch on computers from hosts list (see --hosts)

 * repeat :     repeat CMD or SCRIPT tasks (see --tasks)

 * hierarchic : handle hierarchic tasks from file
   File format :                   Task dependencies :

   all: a b FILE                           all
   a:                                      / \\
     echo "task a"                        a   b
   b:
     echo "task b"


   Here FILE follows hierarchical task format or a classical task format.

--------------------------------------------------------------------------------
'''

def tasks_help(cmd):
    print '''
--------------------------------------------------------------------------------
 * Several tasks dispatch on several hosts
 --tasks FILE
    $ cat FILE
    sh cmd1.sh --opt1
    python cmd2.py --opt2 arg2
    $ %s --tasks FILE

 * One task repeated on each hosts
 --tasks SCRIPT --mode repeat
    $ cat SCRIPT
    #!/usr/bin/env bash
    echo "my script on $HOSTNAME";
    cat /proc/cpuinfo;

 * Several tasks dispatch on several hosts
 --tasks CMD
    $ %s --tasks 'sh cmd1.sh; python cmd2.py'

 * One task repeated on each hosts
 --tasks CMD --mode repeat
    $ %s --tasks 'echo $HOSTNAME'

 * Empty task (or empty file task) : only probe hosts
 --tasks ''
--------------------------------------------------------------------------------
''' % (cmd, cmd, cmd)

def hosts_help(cmd):
    print '''
--------------------------------------------------------------------------------
* Several hosts from a file
 --hosts FILE
    $ cat file
    file1 file2 file3
    $ %s --hosts file

* Several hosts from command line
 --hosts HOSTS_LIST
    $ %s --hosts host1,host2,host3

* One host from command line
    $ %s --hosts host1,
--------------------------------------------------------------------------------
''' % (cmd, cmd, cmd)
    sys.exit(1)

def timeslot_help(cmd):
    print '''
--------------------------------------------------------------------------------
 * Tasks are started beetween 2 times
 --timeslot H1:M1:S1-H2:M2:S2

 * Tasks are started before one time
 --timeslot -H

 * Tasks are started after one time
 --timeslot H-

 * Complex timeslots (list of timeslots)
 --timeslot -H1,H2-H3,H4-

 * No restriction
 --timeslot allday
 --timeslot -
 --timeslot 0:00:00,24:00:00

 * Shortcuts for current timeslots :
 --timeslot night :         00:00:00-08:00:00,20:00:00-24:00:00
 --timeslot midi :          12:00:00-14:00:00
 --timeslot allday :        00:00:00-24:00:00
 --timeslot morning :       08:00:00-12:00:00
 --timeslot afternoon :     12:00:00-20:00:00
--------------------------------------------------------------------------------
'''



def read_hosts(hosts):
    if hosts.find(',') == -1:
        try:
            fd = open(hosts, 'r')
        except exceptions.IOError:
            msg.error("host file '%s' can not be opened" % hosts)
        else:
            hosts = ' '.join(fd.readlines()).replace('\n', ' ')
            fd.close()
    hosts_list = re.split('[\s,]+', hosts)
    if hosts_list[-1] == '': del hosts_list[-1]
    return hosts_list


def read_hierarchic_tasks(tasks_file):
    def read_file_list(file):
        fd = open(file, 'r')
        lines = fd.readlines()
        tasks = TaskList(lines)
        fd.close()
        return tasks

    def read_deps_tokens(line):
        # task : dep1 dep2 dep3
        taskre = re.compile('[ \t]*:[ \t]*')
        try:
            task, deps = re.split(taskre, line)
        except ValueError:
            return None
        deps = deps.rstrip('\n')
        depsre = re.compile('[\t ]+')
        deps = re.split(depsre, deps)
        if deps == ['']: deps = []
        return task, deps

    def read_file_hier(file, tasks_dic):
        DEPS_STATE, TASKS_STATE = 0, 1
        fd = open(file, 'r')
        lines = fd.readlines()
        state = DEPS_STATE
        tasks = []
        task = None
        deps = None
        for l in lines:
            blankre = re.compile('^[ \t]*$')
            if blankre.match(l): continue
            depsre = re.compile('[a-zA-Z0-9_]*[ \t]*:[ \t]*')
            if depsre.match(l):
                state = DEPS_STATE
            if state == DEPS_STATE:
                if task: tasks_dic[task] = deps, tasks
                tokens = read_deps_tokens(l)
                state = TASKS_STATE
                tasks = []
                if tokens is None:
                    task, deps = "all", []
                    tasks.append(l)
                else:    task, deps = tokens
            elif state == TASKS_STATE:
                tasks.append(l)
        if task is not None: tasks_dic[task] = deps, tasks
        fd.close()

    def create_hietask(file):
        tasks_dic = {}
        read_file_hier(file, tasks_dic)
        if len(tasks_dic) == 0:
            msg.error("task file is empty")
            sys.exit(1)
        for v, (deps, tasks) in tasks_dic.items():
            for d in deps:
                if tasks_dic.has_key(d): continue
                try:
                    hietask = create_hietask(d)
                    tasks_dic[d] = hietask
                except exceptions.IOError:
                    e = "in file '" + file + \
                        "', rule or file '" + d + \
                        "' does not exist"
                    msg.error(e)
                    sys.exit(1)
        return TaskHierarchical(file, tasks_dic)

    try:
        return create_hietask(tasks_file)
    except exceptions.IOError:
        msg.error("file '%s' does not exist.")
        sys.exit(1)


def read_tasks(tasks, mode):
    if mode in ['hierarchical', 'hie']:
        return read_hierarchic_tasks(tasks)
    if mode == 'dispatch' and isinstance(tasks, list):
        rtasks = TaskList(tasks)
    else:
        try:
            # SCRIPT / BATCH
            fd = open(tasks, 'r')
            lines = fd.readlines()
            if mode == 'dispatch': rtasks = TaskList(lines)
            elif mode == 'repeat': rtasks = Task(';'.join(lines))
            fd.close()
        except exceptions.IOError:
            if mode == 'dispatch': rtasks = TaskList(tasks.split(';'))
            elif mode == 'repeat' : rtasks = Task(tasks)
    return rtasks


def read_timeslot(timeslot):
    if timeslot in ( '-', 'allday' ):
        return TimeSlot(0, 86400) # [0:24h]
    elif timeslot == 'morning':
        return TimeSlot(0, 43200) # [0:12h]
    elif timeslot == 'afternoon':
        return TimeSlot(43200, 86400) # [12h:24h]
    elif timeslot == 'midi':
        return TimeSlot(43200, 50400) # [12h:14h]
    elif timeslot == 'night':
        # [0:8h] + [20h:24h] = [20h:8h]
        return TimeSlotList([TimeSlot(0, 28800),
                TimeSlot(72000, 86400)])

        timeslot_list = timeslot.split(',')

    tsl = []
    for one_timeslot in timeslot_list:
        try:
            t1, t2 = one_timeslot.split('-')
        except ValueError:
            msg.error("'%s' : bad timeslot format" % one_timeslot)
            sys.exit(1)
        seconds = []
        for t in [t1, t2]:
            try:
                h, m, s = t.split(':')
            except ValueError:
                msg.error("'%s' : bad hour format" % t)
                sys.exit(1)
            h, m, s = int(h), int(m), int(s)
            seconds.append(s + 60 * (m + 60 * h))
        tsl.append(TimeSlot(seconds[0], seconds[1]))
    if len(timeslot_list) > 1:
        ts = TimeSlotList(tsl)
    else:    ts = tsl[0]
    return ts

def parse_options(parser):
    morehelpmap = {'hosts' : hosts_help, 'tasks' : tasks_help,
        'mode' : mode_help, 'broken' : broken_help, 'log' : log_help,
        'timeslot' : timeslot_help}
    (options, args) = parser.parse_args(sys.argv)
    error = False
    error_msg = []

    if options.morehelp != None:
        try:
            morehelpmap[options.morehelp](sys.argv[0])
            sys.exit(1)
        except KeyError:
            msg.error("No help for '%s'.\n" % options.morehelp)
            msg.write_list([' ', ('*', 'green'),
                ' List of available full helps :\n   '])
            for k in morehelpmap.keys(): msg.write('%s, ' % k)
            msg.write('\n')
            sys.exit(1)
    mandatory_options = {'hosts' : options.hosts}
    timeslot = read_timeslot(options.timeslot)
    if None in mandatory_options.values():
        opt = [n for n, o in mandatory_options.items() if o == None]
        error_msg.append('missing options : %s' % opt)
        error = True
    if not options.mode in ["dispatch", "repeat", "hie", "hierarchical"]:
        error_msg.append("'%s' : unknown task manager mode." % \
                            options.mode)
        error = True
    if error:
        parser.print_help()
        for m in error_msg: msg.error(m)
        sys.exit(1)
    hosts_list = read_hosts(options.hosts)
    if options.tasks:
        tasks_list = read_tasks(options.tasks, options.mode)
    else:    tasks_list = None
    return options, timeslot, hosts_list, tasks_list


class TimeSlot(object):
    '''
    TimeSlot(start, end)

    Define a contiguous timeslot.

    start, end : in second since day beggining.
    '''
    def __init__(self, start, end):
        if start == '':
            self._start = 0
        else:    self._start = start
        if end == '':
            self._end = 86400 # 24h in seconds
        else:    self._end = end

    def is_inside(self, time):
        return time > self._start and time < self._end

    def is_inside_now(self):
        now = numpy.array(time.localtime())
        h, m, s = now[3:6]
        t = s + 60 * (m + 60 * h)
        return self.is_inside(t)


class TimeSlotList(TimeSlot):
    '''
    TimeSlotList(list)

    Define uncontiguous timslots.

    list :  list of timeslots.
    '''
    def __init__(self, list):
        self._list = list

    def is_inside(self, time):
        a = numpy.array([l.is_inside(time) for l in self._list])
        return a.any()


class Task(object):
    '''
    Only one task that will be computed on only one host.
    '''
    def __init__(self, task):
        self._task = task

    def __len__(self):
        return 1

    def get(self):
        return self._task

    def __repr__(self):
        return self._task

class TaskList(Task):
    '''
    List of independent tasks. Each one can be computed on a different task.
    '''
    def __init__(self, tasks):
        self._tasks = tasks

    def __len__(self):
        return len(self._tasks)

    def next(self):
        t = self._tasks.pop()
        return Task(t)

    def append(self, task):
        self._tasks.append(task)

        def __repr__(self):
                return str(self._tasks)

    def __repr__(self):
        return str(self._tasks)


class TaskHierarchical(Task):
    '''
    Hiearchic dependencies of TaskList.
    '''
#FIXME need more doc
    def __init__(self, rule, tasks_dic):
        self.rule = rule
        self._tasks = tasks_dic

    def init(self):
        self._deps_stack = [0]
        self._rule_stack = ['all']
        msg.write_list(['- ', ('all', 'blue'), ' : %s\n' % \
                ', '.join(self._tasks['all'][0])])

    def __len__(self):
        a = numpy.array([len(t) for t in self._tasks.values()])
        return a.sum()

    def _find_leaf(self):
        if len(self._rule_stack) == 0: return True
        rule = self._rule_stack[-1]
        deps, tasks = self._tasks[rule]
        if deps:
            subrule = deps[self._deps_stack[-1]]
            val = self._tasks[subrule]
            self._rule_stack.append(subrule)
            self._deps_stack.append(0)
            if isinstance(val, TaskHierarchical):
                return False
            else: subdeps, subtasks = val
            msg.write_list(['  ' * len(self._rule_stack), '- ',
                ('%s' % subrule, 'blue'),
                ' : %s\n' % ', '.join(subdeps)])
            return self._find_leaf()
        else:    return True

    def _go_up_for_next_leaf(self):
        while 1:
            if not len(self._rule_stack): break
            rule = self._rule_stack[-1]
            val = self._tasks[rule]
            if isinstance(val, TaskHierarchical):
                self._rule_stack.pop()
                self._deps_stack.pop()
                continue
            deps, tasks = val
            dep_number = self._deps_stack[-1]
            if dep_number >= len(deps):
                if tasks == [] or dep_number > len(deps):
                    self._rule_stack.pop()
                    self._deps_stack.pop()
                    continue
            else:
                self._deps_stack[-1] += 1
                break

    def _pop_subhie(self):
        rule = self._rule_stack[-1]
        self._deps_stack[-1] += 1
        return self._tasks[rule]

    def next(self):
        rule = self._rule_stack[-1]
        val = self._tasks[rule]
        if isinstance(val, TaskHierarchical): val = [], []
        deps, tasks = val

        if self._deps_stack[-1] == 0 and deps != []:
            if not self._find_leaf(): return self._pop_subhie()
            deps, tasks = self._tasks[rule]

        # if at the end of deps :
        if self._deps_stack[-1] >= len(deps):
            # no task or task done -> go up
            if tasks == [] or self._deps_stack[-1] > len(deps):
                self._go_up_for_next_leaf()
                if self._rule_stack == []: return None
                rule = self._rule_stack[-1]
                deps, tasks = self._tasks[rule]
                if self._deps_stack[-1] < len(deps):
                    if not self._find_leaf():
                        return self._pop_subhie()

                elif tasks != []:
                    msg.write_list(['  ' * \
                        len(self._rule_stack), '<- ',
                        ('%s' % rule, 'blue'), '\n'])
                rule = self._rule_stack[-1]
                deps, tasks = self._tasks[rule]
        self._deps_stack[-1] += 1
        return TaskList(list(tasks))


class User(object):
    """Define user launching task and identification process.

    Parameters
    ----------
    name
        username.
    passwd
        user passwd or if None, try to get `~/.ssh/id_dsa` dsa key for key connection.
    keytype
        `rsa` or `dsa`.
    """
    def __init__(self, name=None, passwd=None, keytype=None):
        if name is None:
            name = os.getenv('USER')
        self.name = name
        self.keytype = keytype
        if passwd is None:
            self.passwd = None
            if self.keytype == 'dsa' :
                self.keyfile = os.path.join(
                    os.getenv('HOME'), '.ssh', 'id_dsa')
            elif self.keytype == 'rsa' :
                self.keyfile = os.path.join(
                    os.getenv('HOME'), '.ssh', 'id_rsa')
            else:    self.keyfile = None
        elif passwd in ['GET', '', '-']:
            self.keyfile = None
            self.passwd = getpass.getpass()
        else:
            self.keyfile = None
            self.passwd = passwd

    def key(self):
        if self.keyfile is None: return None
        elif self.keytype == 'dsa':
            key = paramiko.DSSKey(filename=self.keyfile)
        elif self.keytype == 'rsa':
            key = paramiko.RSAKey(filename=self.keyfile)
        return key


class Host(object):
    '''
    Host(name, status)

    name :   hostname.
    status : set host status :
    '''
    #FIXME : add host status list
    def __init__(self, name, status):
        self._name = name
        self._status = status

    def __repr__(self):
        return self._name


class ProbeHost(threading.Thread):
    def __init__(self, hosts_manager, hostname):
        threading.Thread.__init__(self)
        self._hosts_manager = hosts_manager
        self._hostname = hostname

    def run(self):
        up, status = self._hosts_manager.isup(self._hostname)
        host = Host(self._hostname, status)
        self._hosts_manager.update_host_status(host, status)
        self._hosts_manager._hosts_probed_number += 1


class HostsManager(object):
    available_status = 0
    not_available_status = 1
    unknown_host_status = 2
    unknown_status = 3

    def __init__(self, list):
        self._list = list
        self._init_lists()
        self.update_all_hosts()

    def _init_lists(self):
        self._available_list = []
        self._not_available_list = []
        self._unknown_list = []
        self._unknown_status_list = []
        HM = HostsManager
        self.status_lists = {\
            HM.available_status : self._available_list,
            HM.not_available_status : self._not_available_list,
            HM.unknown_host_status : self._unknown_list,
            HM.unknown_status : self._unknown_status_list }

    def update_all_hosts(self):
        self._init_lists()
        self._hosts_probed_number = 0
        for h in self._list:
            ph = ProbeHost(self, h)
            ph.start()
        while 1:
            if self._hosts_probed_number == len(self._list): break
            time.sleep(0.01)
        hosts_lists = {
            'available hosts' :     self._available_list,
            'unavailable hosts' : self._not_available_list,
            'unknown hosts' :       self._unknown_list,
            'unknown status' :      self._unknown_status_list}
        for name, list in hosts_lists.items():
            if len(list) == 0: continue
            msg.write_list([(' * ', 'bold_yellow'), '%s : %s' % \
                            (name, ' ' * (20 - len(name))),
                            ('%d' % len(list), 'cyan'), " ",
                            str(list), '\n'])

    def update_host_status(self, host, status):
        for l in self.status_lists.values():
            try: l.remove(host)
            except ValueError: pass
        self.status_lists[status].append(host)

    def probe(self, hostname):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(3.)
            sock.connect((hostname, 22))
            t = paramiko.Transport(sock)
            return t, self.available_status
        except socket.error, e:
            #print 'hostname', hostname
            #print e
            return None, self.not_available_status
        except socket.gaierror:
            return None, self.unknown_host_status

    def isup(self, hostname):
        t, status = self.probe(hostname)
        if t:
            t.close()
            return True, status
        else:    return False, status


class TasksStarter(threading.Thread):
    def __init__(self, tasks_manager, host, task, time_limit=86400):
        threading.Thread.__init__(self)
        self._tasks_manager = tasks_manager
        self._host = host
        self._task = task
        self._lock = threading.Lock()
        tasks_manager._active_hosts.append(self._host)
        self._killed = False
        self._pty = None
        self._stdout_data = None
        self._start_time = None
        self._status = None
        self._time_limit = time_limit

    def _stop(self, transport, channel):
        if channel: channel.close()
        if transport: transport.close()
        try:
            self._tasks_manager._active_hosts.remove(self._host)
        # task allready finished
        except ValueError: pass

    def _bad_stop(self, transport, channel):
        self._lock.acquire()
        self._stop(transport, channel)
        if not self._killed:
            msg.warning("remove host '%s' from avalaible hosts." % \
                                self._host)
            self._tasks_manager.abnormal_stop(self._task.get())
        self._lock.release()

    def _remote_killed_stop(self, transport, channel):
        self._write_killed_log()
        self._bad_stop(transport, channel)

    def _log_header(self):
        start_time = self._start_time
        stop_time = time.time()
        diff_time = datetime.timedelta(0, int(stop_time - start_time))
        header = "# %s : %s -> %s, on %s\n# %s\n" % (diff_time, \
            time.asctime(time.localtime(start_time)),
            time.asctime(time.localtime(stop_time)),
            self._host, self._task.get().rstrip('\n'))
        header += '# exit status = %s\n' % self._status
        return header

    def _write_killed_log(self):
        header = self._log_header()
        tm = self._tasks_manager
        self._lock.acquire()
        s = self._status - 128
        error = "grid.py error : Killed process by signal %d." % s
        tm._log.write(header + '\n' + error + '\n')
        tm._log.flush()
        self._lock.release()

    def _write_log(self, channel):
        tm = self._tasks_manager
        stdout = [l.strip('\n') for l in channel.makefile('r+')]
        if len(stdout) == 0:
            stdout = [self._stdout_data]
        else:    stdout[0] = self._stdout_data + stdout[0]
        stderr = [l.strip('\n') for l in channel.makefile_stderr('r+')]

        if len(stderr):
            stderr.reverse(); stderr.append('# (stderr)')
            stderr.reverse();
        header = self._log_header()
        self._lock.acquire()
        tm._log.write('\n'.join([header] + stdout + stderr + ['\n']))
        tm._log.flush()
        self._lock.release()

    def _write_broken(self):
        self._lock.acquire()
        tm = self._tasks_manager
        tm._brokenfd.write(self._task.get().strip() + '\n')
        tm._brokenfd.flush()
        self._lock.release()

    def _good_stop(self, transport, channel):
        self._write_log(channel)
        if self._status != 0: self._write_broken()
        self._lock.acquire()
        self._stop(transport, channel)
        self._tasks_manager._free_hosts.append(self._host)
        self._tasks_manager._tasks_finished_number += 1
        self._lock.release()

    def _read_pty(self, chan):
        data = chan.recv(20)
        ind = data.find(' ')
        if ind == -1: ind = len(data)
        self._pty, self._stdout_data = data[:ind], data[ind + 1:]

    def _get_connection_params(self):
        tm = self._tasks_manager
        username = tm._user.name
        key = tm._user.key()
        if key is not None:
            params = {'username' : username, 'pkey' : key}
        else:
            passwd = tm._user.passwd
            params = {'username' : username, 'password' : passwd}
        return params

    def kill(self):
        tm = self._tasks_manager
        try:
            t, status = tm._hosts_manager.probe(self._host._name)
            t.connect(**self._get_connection_params())
            chan = t.open_session()
            cmd = "pkill -t %s" % re.split("/dev/", self._pty)[1]
            chan.exec_command(cmd)
            self._killed = True
            self._bad_stop(t, chan)
        except exceptions.Exception, e:
            msg.warning("can't kill tasks on %s : %s" % \
                            (self._host, e))

    def run(self):
        tm = self._tasks_manager
        hm = tm._hosts_manager
        task = self._task.get()

        # probe host
        hostname = self._host._name
        t, status = hm.probe(hostname)
        if t is None :
            msg.warning('probing %s failed' % hostname)
            #hm.update_host_status(hostname, status)
            self._bad_stop(t, None)
            return

        # get connection
        try:
            t.connect(**self._get_connection_params())
#FIXME : version 1.5.2 (at home) : n'existe pas, verifier a neurospin
#        except paramiko.AuthenticationException:
#            msg.warning('Authentication failed on %s' % hostname)
#            self._bad_stop(t, None)
#            return
        except paramiko.SSHException, e:
            msg.warning('ssh on %s: %s' % (hostname, e))
            self._bad_stop(t, None)
            return
        except exceptions.Exception, e:
            msg.warning('unknown exception (connection) : %s' % e)
            self._bad_stop(t, None)
            return

        # run task
        try:
            chan = t.open_session()
            if chan is None :
                msg.warning('ssh on %s' % hostname)
                self._bad_stop(t, None)
                return
        except paramiko.SSHException, e:
            msg.warning('ssh on %s: %s' % (hostname, e))
            self._bad_stop(t, None)
            return
        except exceptions.Exception, e:
            msg.warning('unknown exception (session opening) : %s' % e)
            self._bad_stop(t, None)
            return
        # associate a tty (pseudo terminal) to connection
        try:
            chan.get_pty()
        except paramiko.SSHException, e:
            msg.warning("can't get tty on %s : %s" % (hostname, e))
            self._bad_stop(t, None)
            return
        #time_limit = "86400" # 24h #FIXME : configurer ce truc
        mem_cmd_bash = "mem=$(cat /proc/meminfo | grep MemTotal " + \
                    "| sed \"s/.*: *//g;s/ kB//g\")"
        mem_cmd_csh = "setenv mem `cat /proc/meminfo | grep MemTotal "+\
                    "| sed \"s/.*: *//g;s/ kB//g\"`"
        limit_cmd_bash = "ulimit -v $mem -t %d" % self._time_limit
        limit_cmd_csh = "limit memoryuse $mem; " + \
                    "limit cputime %d" % self._time_limit
        pre_cmd_bash = "echo BASH;tty;%s;%s;if [ -a ~/.bashrc ]; then source ~/.bashrc;fi;" % \
                    (mem_cmd_bash, limit_cmd_bash)
        pre_cmd_csh = "echo CSH;tty;%s;%s;if [ -a ~/.cshrc ]; then source ~/.cshrc;fi;" % \
                    (mem_cmd_csh, limit_cmd_csh)
        pre_cmd = "(test `basename $SHELL` = 'bash' -o " + \
          "`basename $SHELL` = 'zsh') && eval '%s' || eval '%s';" % \
          (pre_cmd_bash, pre_cmd_csh)
          #print 'pre_cmd:', pre_cmd
        self._start_time = time.time()
        try:
            chan.exec_command(pre_cmd + task)
        except paramiko.SSHException, e:
            msg.warning('ssh on %s: %s' % (hostname, e))
            self._bad_stop(t, chan)
            return
        except exceptions.Exception, e:
            msg.warning('unknown exception : %s' % e)
            self._bad_stop(t, None)
            return

        self._read_pty(chan)
        self._status = chan.recv_exit_status()
        if self._status == -1 or self._status >= 128:
            self._remote_killed_stop(t, chan)
            return
        self._good_stop(t, chan)


class TasksManager(object):
    def __init__(self, timeslot, user, tasks, hosts_manager,
                     log, brokenfd, time_limit=86400):
        self._timeslot = timeslot
        self._user = user
        self._tasks = tasks
        self._hosts_manager = hosts_manager
        self._log = log
        self._brokenfd = brokenfd
        self._free_hosts = list(self._hosts_manager._available_list)
        self._active_hosts = []
        self._time_limit = time_limit

    def print_status(self, n, size):
        msg.write_list(['\r ', ('%d' % n, 'red'),
                '/', ('%d' % size, 'red')])
        if n == size: msg.write('\n')
        sys.stdout.flush()

    def wait_to_be_ready(self):
        we_have_waited = False
        if not self._timeslot.is_inside_now():
            we_have_waited = True
            now = time.asctime(time.localtime())
            msg.info('all tasks stopped : %s' % now)
        while not self._timeslot.is_inside_now():
            time.sleep(60)
        if we_have_waited:
            now = time.asctime(time.localtime())
            msg.info('start all remaining tasks : %s' % now)

    def abnormal_stop(self, task):
        self._tasks.append(task)


class DispatchedTasksManager(TasksManager):
    def __init__(self, *args, **kwargs):
        TasksManager.__init__(self, *args, **kwargs)
        self._tasks_finished_number = 0

    def wait_for_end_or_cmd(self, print_number, tasks_number, cmd):
        while 1:
            try: return cmd()
            except exceptions.IndexError: pass
            if self._tasks_finished_number == tasks_number:
                return None
            if len(self._free_hosts) == 0 and \
                len(self._active_hosts) == 0: return None
            if print_number[0] != self._tasks_finished_number:
                self.print_status(self._tasks_finished_number,
                    tasks_number)
                print_number[0] = self._tasks_finished_number
            time.sleep(0.01)

    def start(self):
        n = len(self._tasks)
        p = 0 # printed tasks number
        host_pop = self._free_hosts.pop
        while self._tasks_finished_number != n:
            host = self.wait_for_end_or_cmd([p], n, host_pop)
            get_next_task = self._tasks.next
            task = self.wait_for_end_or_cmd([p], n, get_next_task)
            if host and task:
                self.wait_to_be_ready()
                ts = TasksStarter(self, host, task,
                                                  self._time_limit)
                ts.start()
            if len(self._free_hosts) == 0 and \
                len(self._active_hosts) == 0: break
            if p != self._tasks_finished_number:
                self.print_status(self._tasks_finished_number,n)
                p = self._tasks_finished_number
            time.sleep(0.01)
        if len(self._free_hosts) == 0 and \
            self._tasks_finished_number != n :
            msg.warning('no more available hosts : sorry :(')


class RepeatedTasksManager(TasksManager):
    def __init__(self, *args, **kwargs):
        TasksManager.__init__(self, *args, **kwargs)
        self._tasks_finished_number = 0

    # FIXME : saute les machines temporairement en rade
    # FIXME : Si une machine est en rade, la tache est ajoutee :
    # self._tasks_manager._tasks.append(task)
    # alors que dans notre cas nous n'avons qu'une seule tache ou le
    # append est imposible. Ce sont les machines qu'il faudrait ajouter
    # a une liste de machine a retester eventuellement.
    def start(self):
        if len(self._tasks) != 1: raise exceptions.RuntimeError
        hosts = self._hosts_manager._available_list
        for host in hosts:
            self.wait_to_be_ready()
            ts = TasksStarter(self, host, self._tasks, self._time_limit)
            ts.start()
        while self._tasks_finished_number != len(hosts):
            self.print_status(self._tasks_finished_number,
                    len(hosts))
            time.sleep(0.01)
        self.print_status(self._tasks_finished_number, len(hosts))

    def abnormal_stop(self, task):
        self._tasks_finished_number += 1

class OneTaskManager(TasksManager):
    def __init__(self, *args, **kwargs):
        TasksManager.__init__(self, *args, **kwargs)
        self._tasks_finished_number = 0

    def start(self):
        if len(self._tasks) != 1: raise exceptions.RuntimeError
        hosts = self._hosts_manager._available_list
        if len(hosts) != 1: raise exceptions.RuntimeError
        self.wait_to_be_ready()
        ts = TasksStarter(self, hosts[0], self._tasks)
        ts.start()
        while len(self._task()) != 0: time.sleep(0.01)
        self.print_status(self._tasks_finished_number, 1)

    def abnormal_stop(self, task):
        pass # if task abnormally failed do nothing

class HierarchicalTasksManager(DispatchedTasksManager):
    def start(self):
        self._tasks.init()
        while 1:
            tasks = self._tasks.next()
            if tasks is None: break
            if isinstance(tasks, TaskHierarchical):
                rule = tasks.rule
                msg.write_list(['--> ', ('file', 'green'),
                    (" : '%s'\n" % rule)])
                htm = HierarchicalTasksManager(self._timeslot,
                    self._user, tasks, self._hosts_manager,
                    self._log, self._brokenfd)
                htm.start()
                msg.write_list(['<-- ', ('file', 'green'),
                    (" : '%s'\n" % rule)])
            else:
                dtm = DispatchedTasksManager(self._timeslot,
                    self._user, tasks, self._hosts_manager,
                    self._log, self._brokenfd)
                dtm.start()

def run_grid(mode, hosts_list, keytype, tasks, timeslot, brokenfile=None,
             logfile=None, user=None, passwd=None, time_limit=86400):
    if 0:
        print 'run_grid ...'
        print 'mode:', mode
        print 'hosts_list:', hosts_list
        print 'keytype:', keytype
        print 'tasks:', tasks
        print 'timeslot:', timeslot
        print 'brokenfile:', brokenfile
        print 'logfile:', logfile
        print 'user:', user
        print 'passwd:', passwd
        print 'time_limit:', time_limit
        print ''
        print ''
    hm = HostsManager(hosts_list)
    if len(hm._available_list) == 0:
        msg.error("All hosts down :(")
        return
    if tasks is None : return
        #print 'run_grid ...'
        #print 'tasks:', tasks
    brokenfd = open(brokenfile or os.devnull, 'w')
    log = open(logfile or os.devnull, 'w')
    user = User(user, passwd, keytype)
    args = (timeslot, user, tasks, hm, log, brokenfd, time_limit)
    if mode == 'dispatch': tm = DispatchedTasksManager(*args)
    elif mode == 'repeat': tm = RepeatedTasksManager(*args)
    elif mode in ['hie', 'hierarchic']:
        tm = HierarchicalTasksManager(*args)
    tm.start()
    brokenfd.close()
    log.close()

def main():
    # options
    parser = create_options(sys.argv)
    options, timeslot, hosts_list, tasks = parse_options(parser)
    run_grid(options.mode, hosts_list, options.keytype, tasks, timeslot,
             options.brokenfile, options.logfile, options.user,
             options.passwd)
    # hm = HostsManager(hosts_list)
    # if len(hm._available_list) == 0:
    #     msg.error("All hosts down :(")
    #     return
    # if tasks is None : return

    # brokenfd = open(options.brokenfile, 'w')
    # log = open(options.logfile, 'w')
    # user = User(options.user, options.passwd, options.keytype)
    # args = (timeslot, user, tasks, hm, log, brokenfd)
    # if options.mode == 'dispatch': tm = DispatchedTasksManager(*args)
    # elif options.mode == 'repeat': tm = RepeatedTasksManager(*args)
    # elif options.mode in ['hie', 'hierarchic']:
    #     tm = HierarchicalTasksManager(*args)
    # tm.start()
    # brokenfd.close()
    # log.close()

def kill_threads():
    print "\n"
    msg.info('kill threads')
    for t in threading.enumerate():
        if isinstance(t, TasksStarter):
            msg.info("kill task on '%s'" % t._host)
            t._write_broken()
            t.kill()
        t._Thread__stop()

def quit(signal, frame):
    kill_threads()
    sys.exit(1)


def main_safe():
    for sig in [signal.SIGQUIT, signal.SIGINT, signal.SIGSEGV]:
        signal.signal(sig, quit)
    try:
        main()
    except KeyboardInterrupt:
        quit(None, None)

if __name__ == "__main__" : main_safe()



# -*- coding: utf-8 -*-

import unittest
import os
import tempfile
import shutil
import time
import logging

from pprint import pformat

import numpy as np
import numpy.testing as npt

import pyhrf.tools as mtools
import pyhrf

from pyhrf.tools import (diagBlock, get_2Dtable_string, do_if_nonexistent_file,
                         peelVolume3D, cartesian, cached_eval, Pipeline,
                         resampleToGrid, set_leaf, get_leaf, tree_rearrange)


class GeometryTest(unittest.TestCase):

    def test_distance(self):
        pass

    def test_convex_hull(self):

        mask = np.array([[0, 1, 1, 1, 1],
                         [0, 0, 0, 1, 1],
                         [1, 0, 0, 1, 1],
                         [1, 1, 1, 1, 1]])

        expected_concex_hull = np.array([[0, 1, 1, 1, 1],
                                         [0, 1, 1, 1, 1],
                                         [1, 1, 1, 1, 1],
                                         [1, 1, 1, 1, 1]])

        chull = mtools.convex_hull_mask(mask)

        npt.assert_equal(chull, expected_concex_hull)


class CropTest(unittest.TestCase):

    def testBasic(self):
        a = np.zeros((4, 4))
        a[1:3, 1:3] = 1


class MiscTest(unittest.TestCase):

    def test_decorator_do_if_file_exist(self):
        from pyhrf.tools import do_if_nonexistent_file

        @do_if_nonexistent_file('fn', 'other_file')
        def perform_task(a, fn, blah, other_file='coco'):
            return True

        f = './my_file.txt'
        # f does not exist yet, expect perform_task not to be executed
        assert perform_task(1, f, 'yoyo')

        # create f -> perform_task should not be executed
        os.system('touch %s' % f)
        os.system('touch coco')
        if perform_task(0, f, 'yoyo') is not None:
            os.system('rm %s coco' % f)
            raise Exception('perform_task should return None when file exists')
        os.system('rm %s coco' % f)

    def test_decorator_do_if_file_exist2(self):

        def foo(a, b, c='4'):
            return True

        r = do_if_nonexistent_file('a')(foo)("1", "2")
        assert r

    def test_decorator_do_if_file_exist_force(self):
        from pyhrf.tools import do_if_nonexistent_file

        @do_if_nonexistent_file('fn', 'other_file', force=True)
        def perform_task(a, fn, blah, other_file='coco'):
            return True

        f = './my_file.txt'
        os.system('touch %s' % f)
        os.system('touch coco')
        # files exist, but force=True -> expect perform_task to be executed
        if not perform_task(0, f, 'yoyo'):
            os.system('rm %s coco' % f)
            raise Exception('perform_task should return None when file exists')
        os.system('rm %s coco' % f)


class PeelVolumeTest(unittest.TestCase):

    def testPeel(self):

        a = np.ones((5, 5, 5), dtype=int)
        peeledA = peelVolume3D(a)
        correctPeeling = np.zeros_like(a)
        correctPeeling[1:-1, 1:-1, 1:-1] = 1
        assert (peeledA == correctPeeling).all()

        m = np.zeros((6, 6, 6), dtype=int)
        m[1:-1, 1:-1, 1:-1] = 1


def foo(a, b, c=1, d=2):
    return a + b + c + d


class CartesianTest(unittest.TestCase):

    def testCartesianBasic(self):

        domains = [(0, 1, 2), ('a', 'b')]
        cartProd = list(cartesian(*domains))
        assert cartProd == [[0, 'a'], [0, 'b'], [1, 'a'], [1, 'b'], [2, 'a'],
                            [2, 'b']]

    def test_cartesian_apply(self):
        from pyhrf.tools import cartesian_apply
        from pyhrf.tools.backports import OrderedDict

        def foo(a, b, c=1, d=2):
            return a + b + c + d

        # OrderDict to keep track of parameter orders in the result
        # arg values must be hashable
        varying_args = OrderedDict([('a', range(2)),
                                    ('b', range(2)),
                                    ('c', range(2))])

        fixed_args = {'d': 10}

        result_tree = cartesian_apply(varying_args, foo,
                                      fixed_args=fixed_args)

        self.assertEqual(result_tree, {0: {0: {0: 10, 1: 11},
                                           1: {0: 11, 1: 12}},
                                       1: {0: {0: 11, 1: 12},
                                           1: {0: 12, 1: 13}}})

    @unittest.skipIf(not mtools.is_importable('joblib'),
                     'joblib (optional dep) is N/A')
    def test_cartesian_apply_parallel(self):
        from pyhrf.tools import cartesian_apply
        from pyhrf.tools.backports import OrderedDict

        # OrderDict to keep track of parameter orders in the result
        # arg values must be hashable
        varying_args = OrderedDict([('a', range(2)),
                                    ('b', range(2)),
                                    ('c', range(2))])

        fixed_args = {'d': 10}

        result_tree = cartesian_apply(varying_args, foo,
                                      fixed_args=fixed_args,
                                      nb_parallel_procs=4)

        self.assertEqual(result_tree, {0: {0: {0: 10, 1: 11},
                                           1: {0: 11, 1: 12}},
                                       1: {0: {0: 11, 1: 12},
                                           1: {0: 12, 1: 13}}})


class DictToStringTest(unittest.TestCase):

    def testBasic(self):
        d = {'k1': 'v1',
             65480: 'v2',
             'k3': 90,
             600: 143,
             }

    def testOnHierachicDict(self):

        d = {'k1': 'v1',
             'd2': {'d2k1': 0, 1: 190, 'd2-3': {'plop': 'plip', 435: 'fr'}},
             'k3': 1098,
             }

    def testOnNumpyArray(self):

        na = np.random.randn(4, 4)
        d = {'k1': 'v1',
             'd2': {'d2k1': 0, 1: 190, 'na': na,
                    'd2-3': {'plop': 'plip', 435: 'fr'}},
             'k3': 1098,
             }

        sd1 = pformat(d)
        original = np.get_printoptions()
        naf = {'precision': 2, 'linewidth': 100}
        np.set_printoptions(**naf)
        sd2 = pformat(d)
        np.set_printoptions(**original)

    def testOnSpmMat(self):
        spmFile = '/export/TMP/thomas/Data/Localizer/bru2698/functional/fMRI/'\
                  'spm_analysis/SPM.mat'
        if os.path.exists(spmFile):
            from scipy.io.mio import loadmat
            spm = loadmat(spmFile)


class TableStringTest(unittest.TestCase):

    def setUp(self):
        nrows = 10
        ncols = 2
        self.rownames = ['row-%d' % i for i in xrange(nrows)]
        self.colnames = ['col-%d' % i for i in xrange(ncols)]
        self.data1D = np.random.randn(nrows)
        self.data2D = np.random.randn(nrows, ncols)
        self.data3D = np.random.randn(nrows, ncols, 7)
        self.data4D = np.random.randn(nrows, ncols, 7, 20)

    def test1Darray(self):
        s = get_2Dtable_string(self.data1D, self.rownames, ['Value'])

    def test2Darray(self):
        s = get_2Dtable_string(self.data2D, self.rownames, self.colnames)

    def test3Darray(self):
        s = get_2Dtable_string(self.data3D, self.rownames, self.colnames)
        if 0:
            print s

    def test4Darray(self):
        s = get_2Dtable_string(self.data3D, self.rownames, self.colnames,
                               precision=2, outline_char='-', line_end='|',
                               line_start='|')
        if 0:
            print ''
            print s

    def test2Darray_latex(self):
        s = get_2Dtable_string(self.data2D, self.rownames, self.colnames,
                               line_end='\\\\', col_sep='&')
        if 0:
            print 's:'
            print s


class DiagBlockTest(unittest.TestCase):

    def testAll2D(self):

        m1 = np.arange(2 * 3).reshape(2, 3)
        m2 = np.arange(2 * 4).reshape(2, 4)
        bm = diagBlock([m1, m2])

    def testFrom1D(self):

        m1 = np.arange(2 * 3)
        bm = diagBlock([m1])

    def testFromNdarray(self):
        m1 = np.arange(2 * 3)
        bm = diagBlock([m1])

    def testRepFrom1D(self):

        m1 = np.arange(2 * 3)
        bm = diagBlock(m1, 5)

    def testRepFrom2D(self):
        m1 = np.arange(2 * 3).reshape(2, 3)
        bm = diagBlock(m1, 2)

    def testRepFromBlocks(self):

        m1 = np.arange(2 * 3).reshape(2, 3)
        m2 = np.arange(2 * 4).reshape(2, 4)

        bm = diagBlock([m1, m2], 2)


class ResampleTest(unittest.TestCase):

    def testResampleToGrid(self):

        import numpy
        size = 10

        x = numpy.concatenate(
            ([0.], numpy.sort(numpy.random.rand(size)), [1.]))
        y = numpy.concatenate(
            ([0.], numpy.sort(numpy.random.rand(size)), [1.]))
        grid = numpy.arange(0, 1., 0.01)
        ny = resampleToGrid(x, y, grid)

    def testLargerTargetGrid(self):

        import numpy
        size = 10

        x = numpy.concatenate(
            ([0.], numpy.sort(numpy.random.rand(size)), [1.]))
        y = numpy.concatenate(
            ([0.], numpy.sort(numpy.random.rand(size)), [1.]))
        grid = numpy.arange(-0.2, 1.1, 0.01)
        ny = resampleToGrid(x, y, grid)


class treeToolsTest(unittest.TestCase):

    def test_set_leaf(self):
        d = {}
        set_leaf(d, ['b1', 'b2', 'b3'], 'theLeaf')

    def test_get_leaf(self):
        d = {}
        set_leaf(d, ['b1', 'b2', 'b3.1'], 'theLeaf1')
        set_leaf(d, ['b1', 'b2', 'b3.2'], 'theLeaf2')
        l1 = get_leaf(d, ['b1', 'b2', 'b3.1'])
        l2 = get_leaf(d, ['b1', 'b2', 'b3.2'])
        return l2

    def test_walk_branches(self):
        d = {}
        set_leaf(d, ['b1', 'b2.1', 'b3.1'], 'theLeaf1')
        set_leaf(d, ['b1', 'b2.1', 'b3.2'], 'theLeaf2')
        set_leaf(d, ['b1', 'b2.2', 'b3.3'], 'theLeaf3')

    def test_stack_trees(self):
        d1 = {}
        set_leaf(d1, ['b1', 'b2.1', 'b3.1'], 'd1Leaf1')
        set_leaf(d1, ['b1', 'b2.1', 'b3.2'], 'd1Leaf2')
        set_leaf(d1, ['b1', 'b2.2', 'b3.3'], 'd1Leaf3')

        d2 = {}
        set_leaf(d2, ['b1', 'b2.1', 'b3.1'], 'd2Leaf1')
        set_leaf(d2, ['b1', 'b2.1', 'b3.2'], 'd2Leaf2')
        set_leaf(d2, ['b1', 'b2.2', 'b3.3'], 'd2Leaf3')

    def test_rearrange(self):

        d1 = {}
        set_leaf(d1, ['1', '2.1', 'o1'], 'c1')
        set_leaf(d1, ['1', '2.1', 'o2'], 'c2')
        set_leaf(d1, ['1', '2.2', 'o1'], 'c3')
        set_leaf(d1, ['1', '2.2', 'o2'], 'c4')

        blabels = ['p1', 'p2', 'outname']

        nblabels = ['outname', 'p1', 'p2']
        tree_rearrange(d1, blabels, nblabels)


def foo_func(a, b):
    return a + b


def slow_func(a, b):
    time.sleep(1)
    return a + b


class CachedEvalTest(unittest.TestCase):

    def setUp(self):
        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.cache_dir = tmpDir

    def tearDown(self):
        shutil.rmtree(self.cache_dir)

    def test_simple(self):
        cached_eval(foo_func, (1, 2), path=self.cache_dir)

    def test_simple_args(self):
        cached_eval(foo_func, {'a': 4, 'b': 6}, path=self.cache_dir)

    def test_slow_func(self):
        t0 = time.time()
        cached_eval(slow_func, {'a': 4, 'b': 6}, path=self.cache_dir)
        delta = time.time() - t0
        t0 = time.time()
        cached_eval(slow_func, {'a': 4, 'b': 6}, path=self.cache_dir)
        delta = time.time() - t0
        assert delta < .1
        t0 = time.time()
        cached_eval(slow_func, {'a': 4, 'b': 8}, path=self.cache_dir)
        delta = time.time() - t0

    def test_code_digest(self):
        t0 = time.time()
        cached_eval(slow_func, {'a': 4, 'b': 8}, digest_code=True,
                    path=self.cache_dir)
        delta = time.time() - t0


def computeB(a, e):
    return a + e


def computeC(a):
    return a ** 2


def computeF(g, e):
    return g / e


def computeD(f, b, c):
    return (f + b) * c


def computeJ(i, l):
    return i ** 3 + l


def computeK(j):
    return j / 2.


def computeL(k):
    return k / 3.


def foo_default_arg(a, d=1):
    return a + d


def foo_a(c=1):
    return c


def foo_multiple_returns(e):
    return e, e / 2


class PipelineTest(unittest.TestCase):

    def setUp(self):

        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.cache_dir = tmpDir

        # Be sure to put scalars in arrays so that they can be referenced
        self.a = np.array([3])
        self.e = np.array([10])
        self.g = np.array([9.3])

        # Good acyclic dependence tree :
        #   g-\.
        #   e--.f
        #    \.  \
        #     .b  \.
        #   a/ \___.d
        #    \.c__/
        self.depTree = {'a': self.a,
                        'e': self.e,
                        'g': self.g,
                        'b': computeB,
                        'c': computeC,
                        'f': computeF,
                        'd': computeD,
                        }

        # Bad cyclic dependence tree :
        #    j.
        #    | \
        # i  . .l
        #  \.k/

        self.i = np.array([34.5])

        self.badDepTree = {'i': self.i,
                           'j': computeJ,
                           'k': computeK,
                           'l': computeL
                           }

    def tearDown(self):
        shutil.rmtree(self.cache_dir)

    def testGoodDepTreeInit(self):

        data = Pipeline(self.depTree)
        assert (data.get_value('b') == self.a + self.e).all()
        assert (data.get_value('c') == self.a ** 2).all()
        assert (data.get_value('d') ==
                (self.g / self.e + self.a + self.e) * self.a ** 2).all()
        assert (data.get_value('f') == self.g / self.e).all()

    def testBadDepTreeInit(self):

        try:
            data = Pipeline(self.badDepTree)
        except Exception:
            pass
        else:
            raise Exception()

    def testRepr(self):
        data = Pipeline(self.depTree)

    def test_func_default_args(self):
        data = Pipeline({'a': foo_a,
                         'b': foo_default_arg,
                         'd': 7})
        data.resolve()

    def test_cached(self):
        try:
            from joblib import Memory
            mem = Memory(self.cache_dir, verbose=0)
            dep_tree = {
                'a': 5,
                'b': 6,
                'c': mem.cache(slow_func),
            }
            data = Pipeline(dep_tree)
            t0 = time.time()
            data.resolve()
            delta = time.time() - t0

            t0 = time.time()
            data.resolve()
            delta = time.time() - t0
            assert delta < .1
        except:
            pass

    def test_multiple_output_values(self):

        data = Pipeline({'e': foo_a,
                         'b': foo_default_arg,
                         ('a', 'd'): foo_multiple_returns})
        data.resolve()



# -*- coding: utf-8 -*-

import os.path as op
import logging

from random import randrange

import numpy as np

from numpy.random import randn
from scipy.signal.signaltools import lfilter

import pyhrf
import pyhrf.paradigm as mpar

from pyhrf.boldsynth.field import genPotts
from pyhrf.tools import buildPolyMat
from pyhrf.graph import graph_from_lattice, kerMask2D_4n, kerMask3D_6n
from pyhrf import paradigm
from pyhrf.boldsynth import hrf as shrf
from pyhrf.tools._io import write_volume
from pyhrf.graph import bfs_set_label
from pyhrf.tools import add_prefix
from pyhrf.ndarray import xndarray, expand_array_in_mask
try:
    from collections import OrderedDict
except ImportError:
    from pyhrf.tools.backports import OrderedDict


logger = logging.getLogger(__name__)


def create_varying_hrf(hrf_duration=25., dt=.5):
    timeaxis = np.arange(0, hrf_duration, dt)
    Picw = randrange(2, 4, 1)
    picp = [5, 1]
    h = shrf.genBezierHRF(timeAxis=timeaxis, pic=picp, picw=Picw)
    return h[1]


def create_hrf(picw, pic, under=2, hrf_duration=25., dt=.5):
    timeaxis = np.arange(0, hrf_duration, dt,)
    Picp = [pic, 1]
    upos = [under, -.2]
    h = shrf.genBezierHRF(timeAxis=timeaxis, pic=Picp, picw=picw, ushoot=upos)

    return h[1]


def create_bold_from_stim_induced_RealNoise(stim_induced_signal, dsf, noise,
                                            drift):
    """
    Downsample stim_induced signal according to downsampling factor 'dsf' and
    add noise and drift (nuisance signals) which has to be at downsampled
    temporal resolution.
    """
    bold = stim_induced_signal[0:-1:dsf, :]
    bold += drift
    TR = bold.shape[0]
    noise = np.reshape(noise[0:TR, :, :, :], bold.shape)
    bold += noise

    return bold


def create_labels_Potts(condition_defs, beta, nb_voxels):
    labels = []
    nb_voxels = nb_voxels ** .5  # assume square shape
    shape = (nb_voxels, nb_voxels)
    mask = np.ones(shape, dtype=int)
    graph = graph_from_lattice(mask, kerMask=kerMask2D_4n, toroidal=True)
    shape = (1, nb_voxels, nb_voxels)
    for c in condition_defs:
        tmp = genPotts(graph, beta, 2)
        tmp = np.reshape(tmp, shape)
        labels.append(tmp)
    return np.array(labels)


def create_3Dlabels_Potts(condition_defs, beta, dims, mask):
    labels = []
    graph = graph_from_lattice(mask, kerMask=kerMask3D_6n, toroidal=True)
    for c in condition_defs:
        tmp = genPotts(graph, beta, 2)
        tmp = np.reshape(tmp, dims)
        labels.append(tmp)
    return np.array(labels)


def create_AR_noise(bold_shape, v_noise, order=2, v_corr=0.1):
    noise = np.random.randn(*bold_shape)
    nbVox = bold_shape[1]
    length = bold_shape[0]
    noiseMean = np.zeros(nbVox, dtype=float)
    noiseVar = v_noise * np.random.rand(nbVox)
    ARorder = order
    noiseARp = v_corr * (np.ones((nbVox, ARorder), dtype=float) +
                         np.random.rand(nbVox, ARorder))
    for v in xrange(nbVox):
        tmp = np.sqrt(noiseVar[v]) * np.random.randn(length) + noiseMean[v]
        ARArray = np.concatenate(([1], -noiseARp[v, :]))
        noise[:, v] = lfilter([1], ARArray, tmp)
    return noise


def create_localizer_paradigm(condition_defs, paradigm_label='av'):
    ons = eval('paradigm.onsets_loc_%s' % paradigm_label)
    cnames = [c.name for c in condition_defs]
    onsets = OrderedDict(zip(cnames, [ons[c] for c in cnames]))
    par = mpar.Paradigm(onsets, sessionDurations=[300.])
    return par


def create_localizer_paradigm_avd(condition_defs):
    ons = mpar.onsets_loc_av_d
    cnames = [c.name for c in condition_defs]
    onsets = OrderedDict(zip(cnames, [ons[c] for c in cnames]))
    par = mpar.Paradigm(onsets, sessionDurations=[300.])
    return par


def create_localizer_paradigm_a(condition_defs):
    ons = mpar.onsets_loc_a
    dur = paradigm.durations_loc_a
    cnames = [c.name for c in condition_defs]
    onsets = OrderedDict(zip(cnames, [ons[c] for c in cnames]))
    durations = OrderedDict(zip(cnames, [dur[c] for c in cnames]))

    p = mpar.Paradigm(onsets, stimDurations=durations, sessionDurations=[300.])
    return p


def create_paradigm_un_evnt(condition_defs):
    ons = mpar.onsets_un_evnt
    cnames = [c.name for c in condition_defs]
    onsets = OrderedDict(zip(cnames, [ons[c] for c in cnames]))
    paradigm = mpar.Paradigm(onsets)
    return paradigm


def create_language_paradigm(condition_defs):
    ons = mpar.onsets_language_sess1
    dur = mpar.durations_language_sess1
    cnames = [c.name for c in condition_defs]
    onsets = OrderedDict(zip(cnames, [ons[c] for c in cnames]))
    durations = OrderedDict(zip(cnames, [dur[c] for c in cnames]))
    paradigm = mpar.Paradigm(onsets, stimDurations=durations)
    return paradigm


def rasterize_paradigm(paradigm, dt, condition_defs):
    """ Return binary sequences of onsets approximated on temporal grid of
    temporal resolution dt, for all conditions. 'paradigm' is expected to be
    an instance of 'pyhrf.paradigm.mpar.Paradigm'
    """
    rparadigm = paradigm.get_rastered(dt)
    return np.vstack([rparadigm[c.name][0] for c in condition_defs])


def load_drawn_labels(name):
    fn = pyhrf.get_data_file_name('simu_labels_%s.png' % name)
    if not op.exists(fn):
        raise Exception('Unknown label map %s (%s)' % (name, fn))

    from scipy.misc import fromimage
    from PIL import Image
    labels_image = Image.open(fn)
    labels_image = labels_image.convert("L")
    labels = fromimage(labels_image)
    return labels[np.newaxis, :, :]


def create_labels_vol(condition_defs):
    """ Create a seet labels from the field "label_map" in *condition_defs*
    Available choices for the field label_map:
    - 'random_small' : binary labels are randomly generated with shape (1,5,5)
    - a tag (str) : corresponds to a png file in pyhrf data files
    - a 3D np containing the labels
    """
    labels = []
    for c in condition_defs:
        if isinstance(c.label_map, np.ndarray):
            labels.append(c.label_map)
            if c.label_map.ndim < 3:
                raise Exception('For cond %s, label map must be 3D, '
                                'got %d dim(s)' % (c.name, c.label_map.ndim))
        elif isinstance(c.label_map, str):
            if c.label_map == 'random_small':
                shape = (1, 5, 5)
                labels.append(np.random.randint(0, 2, shape))
            else:
                labels.append(load_drawn_labels(c.label_map))
        else:
            raise Exception("label_map of type %s not supported"
                            % c.label_map.__class__)

    return np.array(labels)


def create_connected_label_clusters(condition_defs, activ_label_graph):
    graph = activ_label_graph
    labels = np.zeros((len(condition_defs), len(graph)), dtype=int)
    for ic, c in enumerate(condition_defs):
        for center in c.centers:
            bfs_set_label(graph, center, labels[ic, :], 1, c.radius)

    return labels


def flatten_labels_vol(labels_vol):
    vol_mask = np.ones_like(labels_vol[0])
    mask = np.where(vol_mask)
    return np.vstack([labels_vol[ic][mask] for ic in range(labels_vol.shape[0])])


def create_bigaussian_nrls(labels, mean_act, var_act, var_inact):
    """
    Simulate bi-Gaussian NRLs (zero-centered inactive component)
    """
    mask_activ = np.where(labels)
    nrls = np.random.randn(*labels.shape) * var_inact ** .5
    nrls[mask_activ] = np.random.randn(labels.sum()) * var_act ** .5 + mean_act

    return nrls


def create_time_invariant_gaussian_nrls(condition_defs, labels):
    nrls = []
    for ic, c in enumerate(condition_defs):
        labels_c = labels[ic]
        mask_activ = np.where(labels_c)
        nrls_c = randn(labels_c.size) * c.v_inact ** .5 + 0
        nrls_c[mask_activ] = randn(labels_c.sum()) * c.v_act ** .5 + c.m_act
        nrls.append(nrls_c)

    return np.vstack(nrls)


def create_time_invariant_gaussian_brls(condition_defs, labels):
    """ BOLD response levels for ASL """

    nrls = []
    for ic, c in enumerate(condition_defs):
        labels_c = labels[ic]
        mask_activ = np.where(labels_c)
        nrls_c = randn(labels_c.size) * c.bold_v_inact ** .5 + 0
        nrls_c[mask_activ] = randn(labels_c.sum()) * c.bold_v_act ** .5 + \
            c.bold_m_act
        nrls.append(nrls_c)
    return np.vstack(nrls)


def create_time_invariant_gaussian_prls(condition_defs, labels):
    """ Perfusion response levels for ASL """

    nrls = []
    for ic, c in enumerate(condition_defs):
        labels_c = labels[ic]
        mask_activ = np.where(labels_c)
        nrls_c = randn(labels_c.size) * c.perf_v_inact ** .5 + 0
        nrls_c[mask_activ] = randn(labels_c.sum()) * c.perf_v_act ** .5 + \
            c.perf_m_act
        nrls.append(nrls_c)
    return np.vstack(nrls)


def create_gaussian_nrls_sessions_and_mean(nrls, condition_defs, labels,
                                           var_sess):
    '''
    Creation of nrls by session (and by voxel and cond) - for one session n° sess
    The nrls by session vary around an nrl mean (nrls_bar) defined by voxel and cond
    (var_sess corresponds to the variation of session defined nrls around the nrl_bar)
    Here "nrls" is nrls_bar, mean over subjects!
    '''
    nrls_sess = []
    for ic, c in enumerate(condition_defs):
        labels_c = labels[ic]
        nrls_s = randn(labels_c.size) * var_sess ** .5 + nrls[ic]
        nrls_sess.append(nrls_s)
    return np.vstack(nrls_sess)


def createBiGaussCovarNRL(condition_defs, labels, covariance):
    mean_for_corr = [c.m_act for c in condition_defs]
    var = [c.v_act for c in condition_defs]
    sigma = np.diag(var) + covariance
    nrls = create_time_invariant_gaussian_nrls(condition_defs, labels)

    # Definition of covariance structure between conditions,
    # for voxels at the intersection
    mask_intersection = np.prod(labels, axis=0)
    Pos_intersection = np.where(mask_intersection)
    size = len(Pos_intersection[0])
    nrls[:, Pos_intersection[0]] = np.random.multivariate_normal(mean_for_corr,
                                                                 sigma, size).T
    return nrls


def create_gaussian_noise(bold_shape, v_noise, m_noise=0.):
    return np.random.randn(*bold_shape) * v_noise ** .5 + m_noise


def create_gaussian_noise_asl(asl_shape, v_gnoise, m_noise=0.):
    return create_gaussian_noise(asl_shape, v_gnoise, m_noise)


def create_drift_coeffs(bold_shape, drift_order, drift_coeff_var):

    return drift_coeff_var ** .5 * randn(drift_order + 1, bold_shape[1])


def create_polynomial_drift_from_coeffs(bold_shape, tr, drift_order,
                                        drift_coeffs, drift_mean=0.,
                                        drift_amplitude=1.):
    p = buildPolyMat(drift_order, bold_shape[0], tr)
    if 0:
        print '%%%%%%%%%%%%%%%%%%%%%%%%%'
        print 'p:', p
        print '%%%%%%%%%%%%%%%%%%%%%%%%%'
        print 'drift_coeffs:', drift_coeffs.shape
    drift = np.dot(p, drift_coeffs) * drift_amplitude + drift_mean
    logger.info('Drift shape: %s', str(drift.shape))
    return drift


def create_drift_coeffs_asl(asl_shape, drift_order, drift_var):
    return create_drift_coeffs(asl_shape, drift_order, drift_var)


def create_polynomial_drift_from_coeffs_asl(asl_shape, tr, drift_order,
                                            drift_coeffs):
    return create_polynomial_drift_from_coeffs(asl_shape, tr, drift_order,
                                               drift_coeffs)


def create_polynomial_drift(bold_shape, tr, drift_order, drift_var):
    p = buildPolyMat(drift_order, bold_shape[0], tr)
    nvox = bold_shape[1]
    coeff = drift_var ** .5 * randn(p.shape[1], nvox)
    drift = np.dot(p, coeff)
    logger.info('Drift shape: %s', str(drift.shape))
    return drift


def create_null_drift(bold_shape):
    return np.zeros(bold_shape)


def get_bold_shape(stim_induced_signal, dsf):
    return stim_induced_signal[0:-1:dsf, :].shape


def calc_asl_shape(bold_stim_induced, dsf):
    return bold_stim_induced[0:-1:dsf, :].shape


def load_hrf_territories(nb_hrf_territories=0, hrf_territories_name=None):
    if hrf_territories_name is not None:
        fn = pyhrf.get_data_file_name('simu_hrf_%s.png' % hrf_territories_name)
        if not op.exists(fn):  # HACK #TODO improve this
            fn = pyhrf.get_data_file_name(hrf_territories_name)

    else:
        fn = pyhrf.get_data_file_name('simu_hrf_%d_territories.png'
                                      % nb_hrf_territories)
    assert op.exists(fn)
    from scipy.misc import fromimage
    from PIL import Image
    territories = fromimage(Image.open(fn))
    territories = territories[np.newaxis, :, :]
    t = territories[np.where(np.ones_like(territories))]
    labels = np.unique(territories)
    for ilab, lab in enumerate(labels):
        t[np.where(t == lab)] = ilab
    return t


def load_many_hrf_territories(nb_hrf_territories):
    from pyhrf.tools._io import read_volume
    fn = pyhrf.get_data_file_name('simu_hrf_%d_territories.nii'
                                  % nb_hrf_territories)
    assert op.exists(fn)
    territories = read_volume(fn)[0]
    return territories[np.where(np.ones_like(territories))]


def create_hrf_from_territories(hrf_territories, primary_hrfs):

    logger.info('create_hrf_from_territories ...')
    logger.info(' inputs: hrf_territories %s, primary_hrfs (%d,%d)',
                str(hrf_territories.shape), len(primary_hrfs),
                len(primary_hrfs[0][0]))
    assert hrf_territories.ndim == 1
    hrfs = np.zeros((hrf_territories.size, primary_hrfs[0][0].size))
    territories = np.unique(hrf_territories)
    territories.sort()
    if territories.min() == 1:
        territories = territories - 1

    assert territories.min() == 0
    assert territories.max() <= len(primary_hrfs)

    for territory in territories:
        # TODO: test consitency in hrf lengths
        m = np.where(hrf_territories == territory)[0]
        hrfs[m, :] = primary_hrfs[territory][1]
    return hrfs.transpose()


def duplicate_hrf(nb_voxels, primary_hrf):
    """
    Duplicate hrf over all voxels.
    Return an array of shape (nb_voxels, len(hrf))
    """
    hrfs = np.zeros((nb_voxels, len(primary_hrf)))
    hrfs[:, :] = primary_hrf
    return hrfs.transpose()


def create_alpha_for_hrfgroup(alpha_var):
    '''
    Create alpha from a normal distribution, for one subject
    '''
    alpha = np.random.randn(1) * alpha_var ** .5 + 0.
    return alpha[0]


def create_gaussian_hrf_subject(hrf_group, var_subject_hrf, dt, alpha=0.0):
    '''
    Creation of hrf by subject.
    Use group level hrf and variance for each subject
    (var_subjects_hrfs must be a list)
    Simulated hrfs must be smooth enough: correlation between temporal coeffcients
    '''

    nb_coeff = hrf_group.shape[0]

    varR = shrf.genGaussianSmoothHRF(False, nb_coeff, dt, var_subject_hrf)[1]
    Matvar = varR / var_subject_hrf
    D1 = np.eye(nb_coeff, k=1) - np.eye(nb_coeff, k=-1)
    hrf_group = hrf_group + alpha * np.dot(D1, hrf_group)

    hrf_s = np.random.multivariate_normal(hrf_group, np.linalg.inv(Matvar))
    normHRF = (hrf_s ** 2).sum() ** (0.5)
    hrf_s /= normHRF

    return hrf_s


def duplicate_brf(nb_voxels, primary_brf):
    """
    Duplicate brf over all voxels.
    Return an array of shape (nb_voxels, len(brf))
    """
    brfs = np.zeros((nb_voxels, len(primary_brf)))
    brfs[:, :] = primary_brf
    return brfs.transpose()


def duplicate_prf(nb_voxels, primary_prf):
    """
    Duplicate prf over all voxels.
    Return an array of shape (nb_voxels, len(prf))
    """
    prfs = np.zeros((nb_voxels, len(primary_prf)))
    prfs[:, :] = primary_prf
    return prfs.transpose()


def duplicate_noise_var(nb_voxels, v_gnoise):
    """
    Duplicate variance of noise over all voxels.
    Return an array of shape (nb_voxels, var noise)
    """
    vars_n = np.zeros((nb_voxels, np.size(v_gnoise)))
    vars_n[:, :] = v_gnoise
    return vars_n.transpose()[0]


def create_canonical_hrf(hrf_duration=25., dt=.5):
    return shrf.getCanoHRF(hrf_duration, dt)[1]


def create_gsmooth_hrf(hrf_duration=25., dt=.5, order=2, hrf_var=1., zc=True,
                       normalize_hrf=True):
    """ Create a smooth HRF according to the multivariate gaussian prior
    used in JDE
    *hrf_duration* and *dt* are the HRF duration and temporal resolution,
    respectively (in sec.).
    *order* is derivative order constraining the covariance matrix.
    *hrf_var* is the HRF variance.
    *zc* is a flag to impose zeros at the begining and the end of the HRF

    return:
    a np array of HRF coefficients
    """
    tAxis = np.arange(0, hrf_duration + dt, dt)
    prcov = len(tAxis) - 2 * zc
    matQ = shrf.buildFiniteDiffMatrix(order, prcov)
    matQ = np.divide(np.dot(matQ.transpose(), matQ), dt ** (order ** 2))
    matL = np.array(np.transpose(np.linalg.cholesky(matQ / hrf_var)))

    hrf = np.linalg.solve(matL, np.random.randn(prcov, 1))
    if zc:
        hrf = np.concatenate(([[0]], hrf, [[0]]))

    if normalize_hrf:
        hrf /= (hrf ** 2).sum() ** .5

    return hrf.squeeze()


def create_prf(prf_duration=25., dt=.5):
    tAxis = np.arange(0, prf_duration + dt, dt)
    tPic = 5.
    tus = 16.
    prf = shrf.genBezierHRF(timeAxis=tAxis, pic=[tPic, 1], picw=2.5,
                            ushoot=[tus, 0.],
                            normalize=True)[1]
    assert len(prf) == len(tAxis)
    return prf


def create_stim_induced_signal(nrls, rastered_paradigm, hrf, dt):
    """
    Create a stimulus induced signal from neural response levels, paradigm and
    HRF (sum_{m=1}^M a^m X^m h)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given HRF and multiply by nrls. Finally
    compute the sum over conditions.

    Return a bold array of shape (nb scans, nb voxels)
    """

    npos = nrls.shape[1]
    duration_dt = hrf.shape[0] + len(rastered_paradigm[0]) - 1
    bold = np.zeros((duration_dt, npos))
    for ipos in xrange(npos):
        if hrf.ndim == 2:
            h = hrf[:, ipos]
        else:
            h = hrf  # single HRF
        for ic in xrange(nrls.shape[0]):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            activity[np.where(activity == 1)] = nrls[ic, ipos]
            bold[:, ipos] += np.convolve(activity, h)
    return bold


def create_Xh(nrls, rastered_paradigm, hrf, condition_defs,
              dt, hrf_territories=None):
    """
    Retrieve the product X.h
    """
    npos = nrls.shape[1]
    duration_dt = len(hrf[:, 0]) + len(rastered_paradigm[0]) - 1
    Xh = np.zeros((duration_dt, npos))
    for ipos in xrange(npos):
        h = hrf[:, ipos]
        for ic, c in enumerate(condition_defs):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            Xh[:, ipos] += np.convolve(activity, h)

    return Xh


def create_multisess_stim_induced_signal(nrls_session, rastered_paradigm,
                                         hrf, condition_defs,
                                         dt, hrf_territories=None):
    """
    Create a stimulus induced signal from neural response levels, paradigm and
    HRF (sum_{m=1}^M a^m X^m h)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given HRF and multiply by nrls. Finally
    compute the sum over conditions.

    Return a bold array of shape (nb scans, nb voxels)
    """
    npos = nrls_session.shape[1]
    duration_dt = len(hrf[:, 0]) + len(rastered_paradigm[0]) - 1
    bold = np.zeros((duration_dt, npos))
    for ipos in xrange(npos):
        h = hrf[:, ipos]
        for ic, c in enumerate(condition_defs):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            activity[np.where(activity == 1)] = nrls_session[ic, ipos]
            bold[:, ipos] += np.convolve(activity, h)

    return bold

def create_multisess_stim_induced_signal_asl(prls_session, rastered_paradigm,
                                         prf, condition_defs,
                                         dt, hrf_territories=None):
    """
    Create a stimulus induced signal from neural response levels, paradigm and
    HRF (sum_{m=1}^M a^m X^m h)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given HRF and multiply by nrls. Finally
    compute the sum over conditions.

    Return a bold array of shape (nb scans, nb voxels)
    """
    npos = prls_session.shape[1]
    duration_dt = len(prf[:, 0]) + len(rastered_paradigm[0]) - 1
    perf = np.zeros((duration_dt, npos))
    for ipos in xrange(npos):
        h = prf[:, ipos]
        for ic, c in enumerate(condition_defs):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            activity[np.where(activity == 1)] = prls_session[ic, ipos]
            perf[:, ipos] += np.convolve(activity, h)

    return perf

def create_bold_stim_induced_signal(brls, rastered_paradigm, brf, condition_defs,
                                    dt, hrf_territories=None):
    """
    Create a stimulus induced signal for ASL from BOLD response levels,
    paradigm and BRF
    (sum_{m=1}^M a^m X^m h + sum_{m=1}^M c^m W X^m g)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given BRF and multiply by brls.
    Finally compute the sum over conditions.

    Return a asl array of shape (nb scans, nb voxels)
    """
    npos = brls.shape[1]
    duration_dt = len(brf[:, 0]) + len(rastered_paradigm[0]) - 1
    asl = np.zeros((duration_dt, npos))

    for ipos in xrange(npos):
        h = brf[:, ipos]
        for ic, c in enumerate(condition_defs):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            activity[np.where(activity == 1)] = brls[ic, ipos]

            asl[:, ipos] += np.convolve(activity, h)
    return asl


def create_perf_baseline(asl_shape, perf_baseline_var, perf_baseline_mean=0.):
    n = asl_shape[0]
    nvox = asl_shape[1]
    rnd = np.random.randn(nvox) * perf_baseline_var ** .5 + perf_baseline_mean
    #return np.abs(np.tile(rnd, (n, 1)))
    return np.tile(rnd, (n, 1))


def create_perf_stim_induced_signal(prls, rastered_paradigm, prf, condition_defs,
                                    dt, hrf_territories=None):
    """
    Create a stimulus induced signal for ASL from perfusion response levels,
    paradigm and PRF
    (sum_{m=1}^M c^m X^m g)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given PRF and multiply by prls.
    Finally compute the sum over conditions.

    Return a asl array of shape (nb scans, nb voxels)
    """
    npos = prls.shape[1]
    duration_dt = len(prf[:, 0]) + len(rastered_paradigm[0]) - 1
    asl = np.zeros((duration_dt, npos))

    for ipos in xrange(npos):
        g = prf[:, ipos]
        for ic, c in enumerate(condition_defs):
            activity = np.array(rastered_paradigm[ic, :], dtype=float)
            activity[np.where(activity == 1)] = prls[ic, ipos]

            asl[:, ipos] += np.convolve(activity, g)
    return asl


def create_stim_induced_signal_Parsi(nrls, rastered_paradigm, hrf,
                                     condition_defs, dt, w):
    """
    Create a stimulus induced signal from neural response levels, paradigm and
    HRF (sum_{m=1}^M a^m w^m X^m h)
    For each condition, compute the convolution of the paradigm binary sequence
    'rastered_paradigm' with the given HRF and multiply by nrls and W. Finally
    compute the sum over conditions.

    Return a bold array of shape (nb scans, nb voxels)
    """
    npos = nrls.shape[1]
    duration_dt = len(hrf[:, 0]) + len(rastered_paradigm[0]) - 1
    bold = np.zeros((duration_dt, npos))
    for ipos in xrange(npos):
        h = hrf[:, ipos]
        for ic, c in enumerate(condition_defs):
            if w[ic] == 1:
                activity = np.array(rastered_paradigm[ic, :], dtype=float)
            else:
                activity = np.zeros(
                    rastered_paradigm[ic, :].shape, dtype=float)

            activity[np.where(activity == 1)] = nrls[ic, ipos]
            bold[:, ipos] += np.convolve(activity, h)

            return bold


def create_outliers(bold_shape, stim_induced_signal, nb_outliers, outlier_scale=5.):

    nscans, nvoxels = bold_shape
    outliers = np.zeros(bold_shape, dtype=stim_induced_signal.dtype)
    outliers_idx = np.arange(nscans)

    max_value = stim_induced_signal.max() * 5.
    for ivoxel in xrange(nvoxels):
        np.random.shuffle(outliers_idx)
        mask_outliers = outliers_idx[:nb_outliers]
        outliers[mask_outliers, ivoxel] = max_value

    return outliers


def create_bold(stim_induced_signal, dsf, noise, drift=None,
                outliers=None):
    """
    a short-cut for function create_bold_from_stim_induced
    """
    return create_bold_from_stim_induced(stim_induced_signal, dsf, noise, drift,
                                         outliers)


def create_bold_from_stim_induced(stim_induced_signal, dsf, noise, drift=None,
                                  outliers=None):
    """
    Downsample stim_induced signal according to downsampling factor 'dsf' and
    add noise and drift (nuisance signals) which has to be at downsampled
    temporal resolution.
    """
    bold = stim_induced_signal[0:-1:dsf, :].copy()
    logger.info('create_bold_from_stim_induced ...')
    logger.info('bold shape: %s, noise shape: %s', str(bold.shape),
                str(noise.shape))
    if drift is not None:
        logger.info('drift shape: %s', str(drift.shape))
        bold += drift
    if outliers is not None:
        bold += outliers
    bold += noise

    return bold


def create_bold_controlled_variance(stim_induced_signal, alpha, nb_voxels, dsf, nrls,
                                    Xh, drift=None, outliers=None):
    """
    Create BOLD with controlled explained variance
    alpha: percentage of explained variance on total variance
    """
    bold = stim_induced_signal[0:-1:dsf, :].copy()
    bold_shape = bold.shape

    var_stim_induced = stim_induced_signal.var(0)

    if drift is not None:
        var_drift = drift.var(0)
        var_noise = (1 - alpha) / alpha * var_stim_induced - var_drift
        vars_n = np.zeros((nb_voxels, np.size(var_noise)))
        vars_n[:, :] = var_noise
        vars_noise = vars_n.transpose()[0]
        noise = np.random.randn(*bold_shape) * vars_noise ** .5
        bold += drift
    else:
        var_noise = (1 - alpha) / alpha * var_stim_induced
        vars_n = np.zeros((nb_voxels, np.size(var_noise)))
        vars_n[:, :] = var_noise
        vars_noise = vars_n.transpose()[0]
        noise = np.random.randn(*bold_shape) * vars_noise ** .5

    if outliers is not None:
        bold += outliers

    bold += noise

    return bold


def build_ctrl_tag_matrix(asl_shape):
    units = np.ones(asl_shape[0]) * -1
    units[::2] = 1
    return np.diag(units)


def create_asl_from_stim_induced(bold_stim_induced, perf_stim_induced,
                                 ctrl_tag_mat, dsf,
                                 perf_baseline, noise, drift=None, outliers=None):
    """
    Downsample stim_induced signal according to downsampling factor 'dsf' and
    add noise and drift (nuisance signals) which has to be at downsampled
    temporal resolution.
    """
    bold = bold_stim_induced[0:-1:dsf, :].copy()
    perf = np.dot(ctrl_tag_mat, (perf_stim_induced[0:-1:dsf, :].copy() +
                                 perf_baseline))

    logger.info('create_asl_from_stim_induced ...')
    logger.info('bold shape: %s, perf_shape: %s, noise shape: %s, '
                'drift shape: %s', str(bold.shape), str(perf.shape),
                str(noise.shape), str(drift.shape))

    asl = bold + perf
    if drift is not None:
        asl += drift
    if outliers is not None:
        asl += outliers
    asl += noise

    return asl


def save_simulation(simulation, output_dir):
    """ short-hand for simulation_save_vol_outputs
    """
    simulation_save_vol_outputs(simulation, output_dir)


def simulation_save_vol_outputs(simulation, output_dir, bold_3D_vols_dir=None,
                                simulation_graph_output=None, prefix=None,
                                vol_meta=None):
    """ simulation_graph_output : None, 'simple', 'thumbnails' #TODO
    """

    if simulation.has_key('paradigm'):
        fn = add_prefix(op.join(output_dir, 'paradigm.csv'), prefix)
        simulation['paradigm'].save_csv(fn)

    # Save all volumes in nifti format:
    if simulation.has_key('labels_vol'):
        mask_vol = np.ones_like(simulation['labels_vol'][0])
    elif simulation.has_key('mask'):
        mask_vol = simulation.get('mask', None)
    elif simulation.has_key('labels'):
        mask_vol = np.ones_like(simulation['labels'][0])
    else:
        raise Exception('Dunno where to get mask')

    logger.info('Vol mask of shape %s', str(mask_vol.shape))

    fn_mask = add_prefix(op.join(output_dir, 'mask.nii'), prefix)
    write_volume(mask_vol.astype(np.int32), fn_mask, vol_meta)

    if simulation.has_key('hrf_territories'):
        fn_h_territories = add_prefix(op.join(output_dir, 'hrf_territories.nii'),
                                      prefix)

        ht = expand_array_in_mask(simulation['hrf_territories'] + 1, mask_vol)
        write_volume(ht, fn_h_territories, vol_meta)

    if simulation.has_key('hrf'):
        from pyhrf.ndarray import MRI3Daxes
        fn_hrf = add_prefix(op.join(output_dir, 'hrf.nii'), prefix)
        logger.info('hrf flat shape %s', str(simulation['hrf'].shape))
        if simulation['hrf'].ndim == 1:
            hrf = (np.ones(mask_vol.size) * simulation['hrf'][:, np.newaxis])
        else:
            hrf = simulation['hrf']

        hrfs_vol = expand_array_in_mask(hrf, mask_vol, flat_axis=1)
        dt = simulation['dt']
        chrfs = xndarray(hrfs_vol, axes_names=['time', ] + MRI3Daxes,
                         axes_domains={'time': np.arange(hrfs_vol.shape[0]) * dt})
        chrfs.save(fn_hrf, vol_meta)

        ttp_vol = hrfs_vol.argmax(0)
        fn_ttp = add_prefix(op.join(output_dir, 'ttp.nii'), prefix)
        write_volume(ttp_vol, fn_ttp, vol_meta)

    if simulation.has_key('brf'):
        from pyhrf.ndarray import MRI3Daxes
        fn_brf = add_prefix(op.join(output_dir, 'brf.nii'), prefix)
        logger.info('brf flat shape %s', str(simulation['brf'].shape))
        brfs_vol = expand_array_in_mask(
            simulation['brf'], mask_vol, flat_axis=1)
        dt = simulation['dt']
        cbrfs = xndarray(brfs_vol, axes_names=['time', ] + MRI3Daxes,
                         axes_domains={'time': np.arange(brfs_vol.shape[0]) * dt})
        cbrfs.save(fn_brf, vol_meta)

    if simulation.has_key('prf'):
        from pyhrf.ndarray import MRI3Daxes
        fn_brf = add_prefix(op.join(output_dir, 'prf.nii'), prefix)
        logger.info('prf flat shape %s', str(simulation['prf'].shape))
        brfs_vol = expand_array_in_mask(
            simulation['prf'], mask_vol, flat_axis=1)
        dt = simulation['dt']
        cbrfs = xndarray(brfs_vol, axes_names=['time', ] + MRI3Daxes,
                         axes_domains={'time': np.arange(brfs_vol.shape[0]) * dt})
        cbrfs.save(fn_brf, vol_meta)

    if simulation.has_key('drift'):
        fn_drift = add_prefix(op.join(output_dir, 'drift.nii'), prefix)
        logger.info('drift flat shape %s', str(simulation['drift'].shape))
        drift_vol = expand_array_in_mask(simulation['drift'], mask_vol,
                                         flat_axis=1)
        write_volume(np.rollaxis(drift_vol, 0, 4), fn_drift)

    if simulation.has_key('drift_coeffs'):
        fn_drift = add_prefix(op.join(output_dir, 'drift_coeffs.nii'), prefix)
        logger.info(
            'drift flat shape %s', str(simulation['drift_coeffs'].shape))
        drift_vol = expand_array_in_mask(simulation['drift'], mask_vol,
                                         flat_axis=1)
        write_volume(np.rollaxis(drift_vol, 0, 4), fn_drift)

    if simulation.has_key('noise'):
        fn_noise = add_prefix(op.join(output_dir, 'noise.nii'), prefix)
        logger.info('noise flat shape %s', str(simulation['noise'].shape))
        noise_vol = expand_array_in_mask(simulation['noise'], mask_vol,
                                         flat_axis=1)
        write_volume(np.rollaxis(noise_vol, 0, 4), fn_noise, vol_meta)

        fn_noise = add_prefix(op.join(output_dir, 'noise_emp_var.nii'), prefix)
        noise_vol = expand_array_in_mask(simulation['noise'].var(0), mask_vol)
        write_volume(noise_vol, fn_noise, vol_meta)

    if simulation.has_key('noise_var'):
        fn_noise_var = add_prefix(op.join(output_dir, 'noise_var.nii'), prefix)
        logger.info(
            'noise_var flat shape %s', str(simulation['noise_var'].shape))
        noise_var_vol = expand_array_in_mask(simulation['noise_var'], mask_vol)
        write_volume(noise_var_vol, fn_noise_var, vol_meta)

    if simulation.has_key('stim_induced_signal'):
        fn_stim_induced = add_prefix(op.join(output_dir, 'stim_induced.nii'),
                                     prefix)
        logger.info('stim_induced flat shape %s',
                    str(simulation['stim_induced_signal'].shape))
        stim_induced_vol = expand_array_in_mask(simulation['stim_induced_signal'],
                                                mask_vol, flat_axis=1)
        write_volume(np.rollaxis(stim_induced_vol, 0, 4), fn_stim_induced)

    if simulation.has_key('perf_stim_induced'):
        fn_stim_induced = add_prefix(op.join(output_dir, 'perf_stim_induced.nii'),
                                     prefix)
        logger.info('asl_stim_induced flat shape %s',
                    str(simulation['perf_stim_induced'].shape))
        stim_induced_vol = expand_array_in_mask(simulation['perf_stim_induced'],
                                                mask_vol, flat_axis=1)
        write_volume(np.rollaxis(stim_induced_vol, 0, 4), fn_stim_induced)

        fn_stim_induced = add_prefix(op.join(output_dir,
                                             'perf_stim_induced_ct.nii'),
                                     prefix)
        logger.info('asl_stim_induced flat shape %s',
                    str(simulation['perf_stim_induced'].shape))

        dsf = simulation['dsf']
        perf = np.dot(simulation['ctrl_tag_mat'],
                      simulation['perf_stim_induced'][0:-1:dsf])
        stim_induced_vol = expand_array_in_mask(perf, mask_vol, flat_axis=1)
        write_volume(np.rollaxis(stim_induced_vol, 0, 4), fn_stim_induced)

    if simulation.has_key('perf_baseline'):
        fn = add_prefix(op.join(output_dir, 'perf_baseline.nii'), prefix)
        pb = np.zeros_like(simulation['bold']) + simulation['perf_baseline']
        write_volume(expand_array_in_mask(pb[0], mask_vol), fn, vol_meta)

    if simulation.has_key('bold_stim_induced'):
        fn_stim_induced = add_prefix(op.join(output_dir, 'bold_stim_induced.nii'),
                                     prefix)
        logger.info('asl_stim_induced flat shape %s',
                    str(simulation['bold_stim_induced'].shape))
        stim_induced_vol = expand_array_in_mask(simulation['bold_stim_induced'],
                                                mask_vol, flat_axis=1)
        write_volume(np.rollaxis(stim_induced_vol, 0, 4), fn_stim_induced)

    m = np.where(mask_vol)
    labels_and_mask = mask_vol.copy()[m]

    for ic in xrange(simulation['labels'].shape[0]):
        if simulation.has_key('condition_defs'):
            c_name = simulation['condition_defs'][ic].name
        else:
            c_name = 'cond%d' % ic
        fn_labels = add_prefix(op.join(output_dir, 'labels_%s.nii' % c_name),
                               prefix)
        if simulation.has_key('labels'):
            labels_c = simulation['labels'][ic]
            labels_and_mask[np.where(labels_c)] = ic + 2
            write_volume(expand_array_in_mask(labels_c, mask_vol).astype(np.int32),
                         fn_labels, vol_meta)
        elif simulation.has_key('labels_vol'):
            labels_c = simulation['labels_vol'][ic]
            labels_and_mask[np.where(labels_c[m])] = ic + 2
            write_volume(labels_c.astype(np.int32), fn_labels, vol_meta)

        if simulation.has_key('nrls'):
            nrls_c = simulation['nrls'][ic]
            fn = add_prefix(
                op.join(output_dir, 'nrls_%s.nii' % c_name), prefix)
            write_volume(expand_array_in_mask(nrls_c, mask_vol), fn, vol_meta)
        if simulation.has_key('nrls_session'):
            nrls_session_c = simulation['nrls_session'][ic]
            fn = add_prefix(op.join(output_dir, 'nrls_session_%s.nii'
                                    % (c_name)), prefix)
            write_volume(expand_array_in_mask(nrls_session_c, mask_vol),
                         fn, vol_meta)

        if simulation.has_key('brls'):
            brls_c = simulation['brls'][ic]
            fn = add_prefix(
                op.join(output_dir, 'brls_%s.nii' % c_name), prefix)
            write_volume(expand_array_in_mask(brls_c, mask_vol), fn, vol_meta)
        if simulation.has_key('prls'):
            prls_c = simulation['prls'][ic]
            fn = add_prefix(
                op.join(output_dir, 'prls_%s.nii' % c_name), prefix)
            write_volume(expand_array_in_mask(prls_c, mask_vol), fn, vol_meta)

        if simulation.has_key('neural_efficacies'):
            ne_c = simulation['neural_efficacies'][ic]
            fn = add_prefix(op.join(output_dir, 'neural_efficacies_%s.nii'
                                    % c_name), prefix)
            write_volume(expand_array_in_mask(ne_c, mask_vol), fn, vol_meta)

    fn_labels_and_mask = add_prefix(op.join(output_dir, 'mask_and_labels.nii'),
                                    prefix)

    write_volume(expand_array_in_mask(labels_and_mask, mask_vol).astype(int),
                 fn_labels_and_mask, vol_meta)

    if simulation.has_key('bold_full_vol') or simulation.has_key('bold'):
        fn = add_prefix(op.join(output_dir, 'bold.nii'), prefix)
        if simulation.has_key('bold_full_vol'):
            bold4D = simulation['bold_full_vol']
        else:
            bold = simulation['bold']
            bold4D = expand_array_in_mask(bold, mask_vol, flat_axis=1)

        write_volume(np.rollaxis(bold4D, 0, 4), fn, vol_meta)

    def save_time_series(k):
        if simulation.has_key(k):
            fn_stim_induced = add_prefix(
                op.join(output_dir, k + '.nii'), prefix)
            logger.info('%s flat shape %s', k, str(simulation[k].shape))
            vol = expand_array_in_mask(simulation[k], mask_vol, flat_axis=1)
            write_volume(np.rollaxis(vol, 0, 4), fn_stim_induced)
    save_time_series('flow_induction')
    save_time_series('cbv')
    save_time_series('hbr')
    save_time_series('bold_stim_induced_rescaled')

    if simulation.has_key('asl'):
        fn = add_prefix(op.join(output_dir, 'asl.nii'), prefix)
        asl4D = expand_array_in_mask(simulation['asl'], mask_vol, flat_axis=1)
        write_volume(np.rollaxis(asl4D, 0, 4), fn, vol_meta)

    if simulation.has_key('outliers'):
        fn = add_prefix(op.join(output_dir, 'outliers.nii'), prefix)
        outliers = expand_array_in_mask(simulation['outliers'], mask_vol,
                                        flat_axis=1)

        write_volume(np.rollaxis(outliers, 0, 4), fn, vol_meta)

    if simulation.has_key('hrf_group'):
        hrfgroup = simulation['hrf_group']
        nb_vox = mask_vol.size
        fn_hrf = add_prefix(op.join(output_dir, 'hrf_group.nii'), prefix)
        logger.info('hrf group shape %s', str(simulation['hrf_group'].shape))
        hrfGd = duplicate_hrf(nb_vox, hrfgroup)
        hrfs_vol = expand_array_in_mask(hrfGd, mask_vol, flat_axis=1)
        dt = simulation['dt']
        chrfs = xndarray(hrfs_vol, axes_names=['time', ] + MRI3Daxes,
                         axes_domains={'time': np.arange(hrfs_vol.shape[0]) * dt})
        chrfs.save(fn_hrf, vol_meta)

    if bold_3D_vols_dir is not None:
        assert op.exists(bold_3D_vols_dir)
        for iscan, bscan in enumerate(bold4D):
            fnout = add_prefix('bold_%06d.nii' % (iscan), prefix)
            write_volume(bscan, op.join(bold_3D_vols_dir, fnout), vol_meta)

from pyhrf.core import Condition
import pyhrf.tools as ptools
import pyhrf.core as pcore


def create_small_bold_simulation(snr="high", output_dir=None, simu_items=None):
    """
    """

    conditions_def = [Condition(name='audio', m_act=3.8, v_act=.5,
                                v_inact=.5,
                                label_map=np.array([[[0, 1], [1, 0]]], dtype=int)),
                      Condition(name='video', m_act=3.8, v_act=.5,
                                v_inact=.5,
                                label_map=np.array([[[1, 0], [1, 0]]], dtype=int))]

    # noive variance
    v_noise = [1.5, .2][snr == "high"]

    simulation_steps = {
        'condition_defs': conditions_def,
        'nb_voxels': 4,
        # Labels
        'labels_vol': create_labels_vol,  # 4D shape
        'labels': flatten_labels_vol,  # 2D shape (flatten spatial axes)
        # Nrls
        'nrls': create_time_invariant_gaussian_nrls,
        'dt': .5,
        'dsf': 2,
        'tr': 1.,
        # Paradigm
        'paradigm': create_localizer_paradigm,
        'rastered_paradigm': rasterize_paradigm,
        # HRF
        'hrf_duration': 25,
        'primary_hrf': create_canonical_hrf,  # one hrf
        'hrf': duplicate_hrf,  # duplicate all HRF along all voxels
        # Stim induced
        'stim_induced_signal': create_stim_induced_signal,
        # Noise
        'v_noise': v_noise,
        'noise': create_gaussian_noise,  # requires bold_shape, v_gnoise
        # Drift
        'drift_order': 4,
        'drift_var': 11.,
        'drift': create_polynomial_drift,
        # Bold
        'bold_shape': get_bold_shape,
        'bold': create_bold_from_stim_induced,
    }

    if simu_items is not None:
        simulation_steps.update(simu_items)

    simu_graph = ptools.Pipeline(simulation_steps)

    simu_graph.resolve()

    if output_dir is not None:
        # Output of the simulation graph:
        simu_graph.save_graph_plot(op.join(output_dir, 'simulation_graph.png'))

    simulation = simu_graph.get_values()

    if output_dir is not None:
        simulation_save_vol_outputs(simulation, output_dir=output_dir,
                                    bold_3D_vols_dir=None,)

    return pcore.FmriData.from_simulation_dict(simulation)
